"""
The :mod:`netharn.examples.cifar` is a clear dataset-specific example of what
netharn is and it's trying to do / not do. For a dataset-agnostic example see
:mod:`netharn.examples.segmentation`.

The basic idea is make an object that inherits from nh.FitHarn. This is our
harness object. It will contain the hyperparameters as well as the learning
state. All the training loop boilerplate has already been written in the parent
class, so all our child class needs to do is: define `prepare_batch` (not
usually needed) and `run_batch`. Code to measure and record performance should
be placed in `on_batch` and `on_epoch`.

The `train` function is our main entry point. It reads parameters from the
command line to override defaults. It then consructs the `HyperParams` object
and constructs an instance of `CIFAR_FitHarn` and calls `harn.run()`.

This begins the training process. At a high level the harness will load the
data using torch DataLoaders, and call `run_batch` when it needs to compute the
model outputs and loss based on the input data. The returned loss is used to
update the model weights if `harn.tag === 'train'`, for validation, test, and
calibration (todo) datasets the loss is simply recorded.

After `run_batch` finishes the `on_batch` function is called, where you can
optionally return a dict of scalars to log as measurements for this batch (note
loss is always recorded, so we need not return it here, but loss components may
be useful). A similar thing happens in `on_epoch`, where you should return
metrics about the entire dataset.

The training harness manages the fit directory structure based on a hash of the
hyperparameters, the creation of algorithm component instance (e.g. model,
optimizer), initializing model weights, restarting from the most recent epoch,
updating the learning rates, various training loop boilerplate details,
checking divergence, reporting progress, handling differences between train,
validation, and test sets. In short, netharn handles the necessary parts and
let the developer focus on the important parts.


References:
    https://github.com/kuangliu/pytorch-cifar


CommandLine:
    python -m netharn.examples.cifar.py --xpu=0 --arch=resnet50
    python -m netharn.examples.cifar.py --xpu=0 --arch=wrn_22 --lr=0.003 --schedule=onecycle --optim=adamw
    python -m netharn.examples.cifar.py --xpu=1,2,3 --arch=wrn_22 --lr=0.003 --schedule=onecycle --optim=adamw --batch_size=1800
    python -m netharn.examples.cifar.py --xpu=1,2 --arch=resnet50 --lr=0.003 --schedule=onecycle --optim=adamw

"""
import sys
from os.path import join
import numpy as np
import ubelt as ub
import torch
import os
import pickle
import netharn as nh
import scriptconfig as scfg
# from netharn.util import layer_rotation


class CIFARConfig(scfg.Config):
    """
    This is the default configuration for running the CIFAR training script.

    Defaults can be modified via passing command line arguments, specifing a
    json configuration file, or using python kwargs.

    Note that scriptconfig is not necessary to use netharn, but it does make it
    easy to specify a default configuration and modify.
    """
    default = {
        'nice': scfg.Value('untitled', help='A human readable tag that is "nice" for humans'),
        'workdir': scfg.Path('~/work/cifar', help='Dump all results in your workdir'),

        'workers': scfg.Value(2, help='number of parallel dataloading jobs'),
        'xpu': scfg.Value('auto', help='See netharn.XPU for details. can be auto/cpu/xpu/cuda0/0,1,2,3)'),

        'dataset': scfg.Value('cifar10', choices=['cifar10', 'cifar100'],
                              help='which cifar network to use'),
        'num_vali': scfg.Value(0, help='number of validation examples'),
        'augment': scfg.Value('baseline', help='an augmentation comma separated list or a code'),

        'arch': scfg.Value('resnet50', help='Network architecture code'),
        'optim': scfg.Value('sgd', help='Weight optimizer. Can be SGD, ADAM, ADAMW, etc..'),

        'input_dims': scfg.Value((32, 32), help='Image size passed to the network'),

        'batch_size': scfg.Value(64, help='number of items per batch'),

        'max_epoch': scfg.Value(350, help='Maximum number of epochs'),
        'patience': scfg.Value(350, help='Maximum "bad" validation epochs before early stopping'),

        'lr': scfg.Value(1e-1, help='Base learning rate'),
        'decay':  scfg.Value(5e-4, help='Base weight decay'),

        'schedule': scfg.Value('step-150-250', help=('Special coercable netharn code. Eg: onecycle50, step50, gamma')),

        'grad_norm_max': scfg.Value(None, help='clip gradients exceeding this value'),
        'warmup_iters': scfg.Value(0, help='number of iterations to warmup learning rate'),

        'init': scfg.Value('noop', help='How to initialized weights. (can be a path to a pretrained model)'),
        'pretrained': scfg.Path(help=('alternative way to specify a path to a pretrained model')),

        'deterministic': scfg.Value(False, help='run deterministically'),
        'seed': scfg.Value(137852547, help='seed for determinism'),
    }

    def normalize(self):
        if self['pretrained'] in ['null', 'None']:
            self['pretrained'] = None

        if self['pretrained'] is not None:
            self['init'] = 'pretrained'


class CIFAR_FitHarn(nh.FitHarn):
    """
    The `FitHarn` class contains a lot of reusable boilerplate. We inherit
    from it and override relevant methods to customize the training procedure
    to our particular problem and dataset.

    Example:
        >>> # xdoctest: +REQUIRES(--download)
        >>> from cifar import *
        >>> harn = setup_harn().initialize()
        >>> batch = harn._demo_batch(0, tag='vali')
        >>> test_metrics = harn._demo_epoch('vali')
    """

    def after_initialize(harn):
        """
        Custom function that performs final initialization steps
        """
        # Our cifar harness will record confusion vectors after every batch.
        # Then, after every epoch we will transform these into quality measures
        # like Accuracy, MCC, AUC, AP, etc...
        harn._accum_confusion_vectors = {
            'y_true': [],
            'y_pred': [],
            'probs': [],
        }

    def run_batch(harn, batch):
        """
        Custom function to compute the output of a batch and its loss.

        The primary purpose of `FitHarn` is to abstract away the boilerplate code
        necessary to run these lines of code.

        After the `FitHarn` prepares the batch it tries to run it.
        In most cases it is best to explicitly override this function to
        describe (in code) exactly how data should be sent to your model and
        exactly how the loss should be computed.

        Returns:
            Tuple[object, Tensor]: This function MUST return:
                (1) the arbitrary output of the model, and
                (2) the loss as a Tensor scalar. The harness will take care
                    of calling `.backwards` on the loss and updating the
                    gradient via the optimizer.
        """
        inputs, labels = ub.take(batch, ['input', 'label'])
        inputs.shape
        outputs = harn.model(inputs)
        loss = harn.criterion(outputs, labels)
        return outputs, loss

    # def backpropogate(harn, bx, batch, loss):
    #     """
    #     Note: this function usually does not need to be overloaded,
    #     but you can if you want to. The actual base implementation is
    #     slightly more nuanced. For details see:
    #     :func:netharn.fit_harn.CoreCallbacks.backpropogate
    #     """
    #     loss.backward()
    #     harn.optimizer.step()
    #     harn.optimizer.zero_grad()

    def on_batch(harn, batch, outputs, loss):
        """
        Custom code executed at the end of each batch.

        This function can optionally return a dictionary containing any scalar
        quality metrics that you wish to log and monitor. (Note these will be
        plotted to tensorboard if that is installed).

        Notes:
            It is best to keep this function small as it is run very often
        """
        y_pred = outputs.data.max(dim=1)[1].cpu().numpy()
        y_true = batch['label'].data.cpu().numpy()

        bx = harn.bxs[harn.current_tag]
        if bx < 3:
            stacked = harn._draw_batch(batch, outputs)
            dpath = ub.ensuredir((harn.train_dpath, 'monitor', harn.current_tag))
            fpath = join(dpath, 'batch_{}_epoch_{}.jpg'.format(bx, harn.epoch))
            import kwimage
            kwimage.imwrite(fpath, stacked)

        probs = outputs.softmax(dim=1).data.cpu().numpy()
        harn._accum_confusion_vectors['y_true'].append(y_true)
        harn._accum_confusion_vectors['y_pred'].append(y_pred)
        harn._accum_confusion_vectors['probs'].append(probs)

    def on_epoch(harn):
        """
        Custom code executed at the end of each epoch.

        This function can optionally return a dictionary containing any scalar
        quality metrics that you wish to log and monitor. (Note these will be
        plotted to tensorboard if that is installed).

        Notes:
            It is ok to do some medium lifting in this function because it is
            run relatively few times.

        Example:
            >>> # xdoctest: +REQUIRES(--download)
            >>> harn = setup_harn().initialize()
            >>> harn._demo_epoch('vali')
        """
        from netharn.metrics import clf_report
        dset = harn.datasets[harn.current_tag]

        probs = np.vstack(harn._accum_confusion_vectors['probs'])
        y_true = np.hstack(harn._accum_confusion_vectors['y_true'])
        y_pred = np.hstack(harn._accum_confusion_vectors['y_pred'])

        # _pred = probs.argmax(axis=1)
        # assert np.all(_pred == y_pred)

        # from netharn.metrics import confusion_vectors
        # cfsn_vecs = confusion_vectors.ConfusionVectors.from_arrays(
        #     true=y_true, pred=y_pred, probs=probs, classes=dset.classes)
        # report = cfsn_vecs.classification_report()
        # combined_report = report['metrics'].loc['combined'].to_dict()

        # ovr_cfsn = cfsn_vecs.binarize_ovr()
        # Compute multiclass metrics (new way!)
        target_names = dset.classes
        ovr_report = clf_report.ovr_classification_report(
            y_true, probs, target_names=target_names, metrics=[
                'auc', 'ap', 'mcc', 'brier'
            ])

        # percent error really isn't a great metric, but its easy and standard.
        errors = (y_true != y_pred)
        acc = 1.0 - errors.mean()
        percent_error = (1.0 - acc) * 100

        metrics_dict = ub.odict()
        metrics_dict['ave_brier'] = ovr_report['ave']['brier']
        metrics_dict['ave_mcc'] = ovr_report['ave']['mcc']
        metrics_dict['ave_auc'] = ovr_report['ave']['auc']
        metrics_dict['ave_ap'] = ovr_report['ave']['ap']
        metrics_dict['percent_error'] = percent_error
        metrics_dict['acc'] = acc

        harn.info(ub.color_text('ACC FOR {!r}: {!r}'.format(harn.current_tag, acc), 'yellow'))

        # Clear confusion vectors accumulator for the next epoch
        harn._accum_confusion_vectors = {
            'y_true': [],
            'y_pred': [],
            'probs': [],
        }
        return metrics_dict

    def _draw_batch(harn, batch, outputs, limit=32):
        """
        Example:
            >>> # xdoctest: +REQUIRES(--download)
            >>> from netharn.examples.cifar import *  # NOQA
            >>> harn = setup_harn().initialize()
            >>> batch = harn._demo_batch(0, tag='test')
            >>> outputs, loss = harn.run_batch(batch)
            >>> stacked = harn._draw_batch(batch, outputs, limit=12)
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> kwplot.imshow(stacked, colorspace='rgb', doclf=True)
            >>> kwplot.show_if_requested()
        """
        import kwimage
        inputs = batch['input'][0:limit].data.cpu().numpy()
        true_cxs = batch['label'].data.cpu().numpy()
        class_probs = outputs.softmax(dim=1).data.cpu().numpy()
        pred_cxs = class_probs.argmax(axis=1)

        dset = harn.datasets[harn.current_tag]
        classes = dset.classes

        todraw = []
        for im, pcx, tcx, probs in zip(inputs, pred_cxs, true_cxs, class_probs):
            im_ = im.transpose(1, 2, 0)

            # Renormalize and resize image for drawing
            min_, max_ = im_.min(), im_.max()
            im_ = ((im_ - min_) / (max_ - min_) * 255).astype(np.uint8)
            im_ = np.ascontiguousarray(im_)
            im_ = kwimage.imresize(im_, dsize=(200, 200),
                                   interpolation='nearest')

            # Draw classification information on the image
            im_ = kwimage.draw_clf_on_image(im_, classes=classes, tcx=tcx,
                                            pcx=pcx, probs=probs)
            todraw.append(im_)

        stacked = kwimage.stack_images_grid(todraw, overlap=-10,
                                            bg_value=(10, 40, 30),
                                            chunksize=8)
        return stacked

    def before_epochs(harn):
        if harn.epoch == 0:
            harn._draw_conv_layers(suffix='_init')

    def after_epochs(harn):
        """
        Callback after all train/vali/test epochs are complete.
        """
        harn._draw_conv_layers()

    def _draw_conv_layers(harn, suffix=''):
        """
        We use this to visualize the first convolutional layer
        """
        import kwplot
        # Visualize the first convolutional layer
        dpath = ub.ensuredir((harn.train_dpath, 'monitor', 'layers'))
        # fig = kwplot.figure(fnum=1)
        for key, layer in nh.util.trainable_layers(harn.model, names=True):
            # Typically the first convolutional layer returned here is the
            # first convolutional layer in the network
            if isinstance(layer, torch.nn.Conv2d):
                if max(layer.kernel_size) > 2:
                    fig = kwplot.plot_convolutional_features(
                        layer, fnum=1, normaxis=0)
                    kwplot.set_figtitle(key, subtitle=str(layer), fig=fig)
                    layer_dpath = ub.ensuredir((dpath, key))
                    fname = 'layer-{}-epoch_{}{}.jpg'.format(
                        key, harn.epoch, suffix)
                    fpath = join(layer_dpath, fname)
                    fig.savefig(fpath)
                    break

            if isinstance(layer, torch.nn.Linear):
                # TODO: visualize the FC layer
                pass


def build_train_augmentors(augment, input_mean):
    from torchvision import transforms

    # Define preprocessing + augmentation strategy
    if isinstance(augment, list):
        augmentors = augment
    elif ',' in augment:
        augmentors = augment.split(',')
    elif augment == 'baseline':
        augmentors = ['crop', 'flip']
    elif augment == 'simple':
        augmentors = ['crop', 'flip', 'gray', 'cutout']
    else:
        raise KeyError(augment)

    pil_augmentors = []
    tensor_augmentors = []

    if 'crop' in augmentors:
        pil_augmentors += [
            transforms.RandomCrop(32, padding=4),
        ]
    if 'flip' in augmentors:
        pil_augmentors += [
            transforms.RandomHorizontalFlip(),
        ]
    if 'gray' in augmentors:
        pil_augmentors += [
            transforms.RandomGrayscale(p=0.1),
        ]
    if 'jitter' in augmentors:
        raise NotImplementedError
        # pil_augmentors += [transforms.RandomChoice([
        #     transforms.ColorJitter(brightness=(0, .01), contrast=(0, .01),
        #                            saturation=(0, .01), hue=(-0.01, 0.01),),
        #     ub.identity,
        # ])]

    if 'cutout' in augmentors:
        def cutout(tensor):
            """
            Ignore:
                tensor = torch.rand(3, 32, 32)
            """
            # This cutout is closer to the definition in the paper
            import kwarray
            rng = kwarray.ensure_rng(None)
            img_h, img_w = tensor.shape[1:]
            p = 0.9
            value = 0
            scale = 0.5
            if rng.rand() < p:
                cx = rng.randint(0, img_w)
                cy = rng.randint(0, img_h)

                w2 = int((img_w * scale) // 2)
                h2 = int((img_h * scale) // 2)
                x1 = max(cx - w2, 0)
                y1 = max(cy - h2, 0)
                x2 = min(cx + w2, img_w)
                y2 = min(cy + h2, img_h)

                sl = (slice(None), slice(y1, y2), slice(x1, x2))
                tensor[sl] = value
            return tensor
        tensor_augmentors += [cutout]

        # tensor_augmentors += [  # Cutout
        #     transforms.RandomErasing(
        #         p=0.5, scale=(0.4, 0.4), ratio=(1.0, 1.0),
        #         value=0, inplace=True),
        # ]
    print('pil_augmentors = {!r}'.format(pil_augmentors))
    print('tensor_augmentors = {!r}'.format(tensor_augmentors))
    return pil_augmentors, tensor_augmentors


def setup_harn():
    """
    This function creates an instance of the custom FitHarness, which involves
    parsing script configuration parameters, creating a custom torch dataset,
    and connecting those data and hyperparameters to the FitHarness.
    """
    import random
    import torchvision
    from torchvision import transforms

    # Create an instance of the CIFAR config and allow command line args
    config = CIFARConfig(cmdline=True)

    xpu = nh.XPU.coerce(config['xpu'])

    # The work directory is where all intermediate results are dumped.
    ub.ensuredir(config['workdir'])

    if config['deterministic']:
        # Take care of random seeding and ensuring appropriate determinisim
        torch.manual_seed((config['seed'] + 0) % int(2 ** 32 - 1))
        random.seed((config['seed'] + 2360097502) % int(2 ** 32 - 1))
        np.random.seed((config['seed'] + 893874269) % int(2 ** 32 - 1))

    if torch.backends.cudnn.enabled:
        # TODO: ensure the CPU mode is also deterministic
        torch.backends.cudnn.deterministic = config['deterministic']

    # A more general system could infer (and cache) this from the data
    input_mean = (0.4914, 0.4822, 0.4465)
    input_std = (0.2023, 0.1994, 0.2010)

    def common_transform(pil_img):
        import kwimage
        hwc255 = np.array(pil_img)
        hwc01 = hwc255.astype(np.float32)
        hwc01 /= 255.0
        if hwc01.shape[0:2] != tuple(config['input_dims']):
            dsize = config['input_dims'][::-1]
            hwc01 = kwimage.imresize(hwc01, dsize=dsize,
                                     interpolation='linear')
        chw01 = torch.from_numpy(hwc01.transpose(2, 0, 1)).contiguous()
        return chw01

    common_transforms = [
        common_transform,
        # transforms.Resize(config['input_dims'], interpolation),
        # transforms.ToTensor(),
        transforms.Normalize(input_mean, input_std, inplace=True),
    ]

    augment = config['augment']
    pil_augmentors, tensor_augmentors = build_train_augmentors(
        augment, input_mean)

    transform_train = transforms.Compose(
        pil_augmentors + common_transforms + tensor_augmentors
    )

    transform_test = transforms.Compose(common_transforms)

    if config['dataset'] == 'cifar10':
        DATASET = torchvision.datasets.CIFAR10
        dset = DATASET(root=config['workdir'], download=True)
        meta_fpath = os.path.join(dset.root, dset.base_folder, 'batches.meta')
        meta_dict = pickle.load(open(meta_fpath, 'rb'))
        classes = meta_dict['label_names']
        # For some reason the torchvision objects dont have the label names
        # in the dataset. But the download directory will have them.
        # classes = [
        #     'airplane', 'automobile', 'bird', 'cat', 'deer', 'dog', 'frog',
        #     'horse', 'ship', 'truck',
        # ]
    elif config['dataset'] == 'cifar100':
        DATASET = torchvision.datasets.CIFAR100
        dset = DATASET(root=config['workdir'], download=True)
        meta_fpath = os.path.join(dset.root, dset.base_folder, 'meta')
        meta_dict = pickle.load(open(meta_fpath, 'rb'))
        classes = meta_dict['fine_label_names']
        # classes = [
        #     'apple', 'aquarium_fish', 'baby', 'bear', 'beaver', 'bed', 'bee',
        #     'beetle', 'bicycle', 'bottle', 'bowl', 'boy', 'bridge', 'bus',
        #     'butterfly', 'camel', 'can', 'castle', 'caterpillar', 'cattle',
        #     'chair', 'chimpanzee', 'clock', 'cloud', 'cockroach', 'couch',
        #     'crab', 'crocodile', 'cup', 'dinosaur', 'dolphin', 'elephant',
        #     'flatfish', 'forest', 'fox', 'girl', 'hamster', 'house',
        #     'kangaroo', 'keyboard', 'lamp', 'lawn_mower', 'leopard', 'lion',
        #     'lizard', 'lobster', 'man', 'maple_tree', 'motorcycle', 'mountain',
        #     'mouse', 'mushroom', 'oak_tree', 'orange', 'orchid', 'otter',
        #     'palm_tree', 'pear', 'pickup_truck', 'pine_tree', 'plain', 'plate',
        #     'poppy', 'porcupine', 'possum', 'rabbit', 'raccoon', 'ray', 'road',
        #     'rocket', 'rose', 'sea', 'seal', 'shark', 'shrew', 'skunk',
        #     'skyscraper', 'snail', 'snake', 'spider', 'squirrel', 'streetcar',
        #     'sunflower', 'sweet_pepper', 'table', 'tank', 'telephone',
        #     'television', 'tiger', 'tractor', 'train', 'trout', 'tulip',
        #     'turtle', 'wardrobe', 'whale', 'willow_tree', 'wolf', 'woman',
        #     'worm']
    else:
        raise KeyError(config['dataset'])

    datasets = {
        'train': DATASET(root=config['workdir'], train=True,
                         transform=transform_train),
        'test': DATASET(root=config['workdir'], train=False,
                        transform=transform_test),
    }

    if config['num_vali']:
        # Create a test train split
        learn = datasets['train']
        learn_copy = DATASET(root=config['workdir'], train=True,
                             transform=transform_test)

        import kwarray
        indices = np.arange(len(learn))
        indices = kwarray.shuffle(indices, rng=0)
        num_vali = config['num_vali']

        datasets['vali'] = torch.utils.data.Subset(learn_copy, indices[:num_vali])
        datasets['train'] = torch.utils.data.Subset(learn, indices[num_vali:])

    # For some reason the torchvision objects do not make the category names
    # easily available. We set them here for ease of use.
    reduction = int(ub.argval('--reduction', default=1))
    for key, dset in datasets.items():
        dset.classes = classes
        if reduction > 1:
            indices = np.arange(len(dset))[::reduction]
            dset = torch.utils.data.Subset(dset, indices)
        dset.classes = classes
        datasets[key] = dset

    loaders = {
        key: torch.utils.data.DataLoader(dset, shuffle=(key == 'train'),
                                         num_workers=config['workers'],
                                         batch_size=config['batch_size'],
                                         pin_memory=True)
        for key, dset in datasets.items()
    }

    if config['workers'] > 0:
        # Solves pytorch deadlock issue #1355.
        import cv2
        cv2.setNumThreads(0)

    if config['optim'] == 'sgd':
        optimizer_ = (torch.optim.SGD, {
            'lr': config['lr'],
            'weight_decay': config['decay'],
            'momentum': 0.9,
            'nesterov': True,
        })
    elif config['optim'] == 'adamw':
        optimizer_ = (nh.optimizers.AdamW, {
            'lr': config['lr'],
            'betas': (0.9, 0.999),
            'weight_decay': config['decay'],
            'amsgrad': False,
        })
    else:
        # The netharn API can construct an optimizer from standard keys in a
        # configuration dictionary. There is a bit of magic involved. Read docs
        # for coerce for more details.
        optimizer_ = nh.api.Optimizer.coerce(config)

    # Choose which network architecture to train
    available_architectures = {
        'densenet121': (nh.models.densenet.DenseNet, {
            'nblocks': [6, 12, 24, 16],
            'growth_rate': 12,
            'reduction': 0.5,
            'num_classes': len(classes),
        }),

        'resnet50': (nh.models.resnet.ResNet, {
            'num_blocks': [3, 4, 6, 3],
            'num_classes': len(classes),
            'block': 'Bottleneck',
        }),

        'dpn26': (nh.models.dual_path_net.DPN, dict(cfg={
            'in_planes': (96, 192, 384, 768),
            'out_planes': (256, 512, 1024, 2048),
            'num_blocks': (2, 2, 2, 2),
            'dense_depth': (16, 32, 24, 128),
            'num_classes': len(classes),
        })),

        'dpn92': (nh.models.dual_path_net.DPN, dict(cfg={
            'in_planes': (96, 192, 384, 768),
            'out_planes': (256, 512, 1024, 2048),
            'num_blocks': (3, 4, 20, 3),
            'dense_depth': (16, 32, 24, 128),
            'num_classes': len(classes),
        })),
    }

    if config['arch'].startswith('wrn'):
        import fastai
        import fastai.vision
        available_architectures['wrn_22'] = (
            fastai.vision.models.WideResNet, dict(
                num_groups=3, N=3, num_classes=len(classes), k=6, drop_p=0.
            )
        )

    if config['arch'].startswith('se_resnet18'):
        from netharn.models import se_resnet
        model = se_resnet.se_resnet18(
            num_classes=len(classes),
        )

    if config['arch'].startswith('se_resnet50'):
        from netharn.models import se_resnet
        model = se_resnet.se_resnet50(
            num_classes=len(classes),
            pretrained=config['init'] == 'cls',
        )

    if config['arch'].startswith('efficientnet'):
        # Directly create the model instance...
        # (as long as it has an `_initkw` attribute)
        from netharn.models import efficientnet

        zero_gamma = False
        if config['init'] == 'cls':
            model_ = efficientnet.EfficientNet.from_pretrained(
                config['arch'], override_params={
                    'classes': classes,
                    'noli': 'mish'
                }, advprop=True)
            print('pretrained cls init')
        else:
            model_ = efficientnet.EfficientNet.from_name(
                config['arch'], override_params={
                    'classes': classes,
                    'noli': 'mish'
                }
            )

        # For efficient nets we need to dramatically reduce the weight decay on
        # the depthwise part of the depthwise separable convolution.  To do
        # this we need to manually construct the param groups for the
        # optimizer.
        model = model_

        params = dict(model.named_parameters())
        key_groups = ub.ddict(list)

        seen_ = set()
        def append_once(group, key):
            if key not in seen_:
                key_groups[group].append(key)
                seen_.add(key)

        if zero_gamma:
            for key, layer in model.trainable_layers(names=True):
                if getattr(layer, '_residual_bn', False):
                    # zero bn after residual layers.
                    layer.weight.data.fill_(0)
                    # dont decay batch norm
                    # append_once('nodecay', key + '.weight')

        for key in params.keys():
            if key.endswith('.bias'):
                append_once('nodecay', key)
            elif 'depthwise_conv' in key:
                append_once('nodecay', key)
            else:
                append_once('default', key)

        named_param_groups = {}
        for group_name, keys in key_groups.items():
            if keys:
                # very important that groups are alway in the same order
                keys = sorted(keys)
                param_group = {
                    'params': list(ub.take(params, keys)),
                }
                named_param_groups[group_name] = param_group

        # Override the default weight decay of chosen groups
        named_param_groups['nodecay']['weight_decay'] = 0

        param_groups = [v for k, v in sorted(named_param_groups.items())]

        optim_cls, optim_kw = optimizer_
        optim = optim_cls(param_groups, **optim_kw)
        optim._initkw = optim_kw
        optimizer_ = optim
    else:
        model_ = available_architectures[config['arch']]

    if config['init'] in ['cls', 'noop']:
        initializer_ = (nh.initializers.NoOp, {})
    if config['init'] == 'kaiming_normal':
        initializer_ = (nh.initializers.KaimingNormal, {'param': 0, 'mode': 'fan_in'})
    else:
        # Note there are lots of different initializers including a special
        # pretrained initializer.
        initializer_ = nh.api.Initializer.coerce(config)

    if config['schedule'] == 'onecycle':
        # TODO: Fast AI params
        # TODO: https://github.com/fastai/fastai/blob/c7df6a5948bdaa474f095bf8a36d75dbc1ee8e6a/fastai/callbacks/one_cycle.py
        scheduler_ = (nh.schedulers.ListedScheduler, {
            'points': {
                'lr': {
                    0   : config['lr'] * 1.00,
                    35  : config['lr'] * 25,
                    70  : config['lr'] / 1000,
                },
                'momentum': {
                    0   : 0.95,
                    35  : 0.85,
                    70  : 0.95,
                },
                # 'weight_decay': {
                #     0: 3e-6,
                # }
            }
        })
    else:
        # The netharn API can construct a scheduler from standard keys in a
        # configuration dictionary. There is a bit of magic involved. Read docs
        # for coerce for more details.
        scheduler_ = nh.api.Scheduler.coerce(config)

    # Notice that arguments to hyperparameters are typically specified as a
    # tuple of (type, Dict), where the dictionary are the keyword arguments
    # that can be used to instantiate an instance of that class. While
    # this may be slightly awkward, it enables netharn to track hyperparameters
    # more effectively. Note that it is possible to simply pass an already
    # constructed instance of a class, but this causes information loss.
    hyper = nh.HyperParams(
        # Datasets must be preconstructed
        datasets=datasets,
        nice=config['nice'],
        # Loader may be preconstructed
        loaders=loaders,
        workdir=config['workdir'],
        xpu=xpu,
        # The 6 major hyper components are best specified as a Tuple[type,
        # dict] However, in recent releases of netharn, these may be
        # initialized manually in certain conditions. See docs for details.
        # TODO: write docs about this.
        model=model_,
        optimizer=optimizer_,
        scheduler=scheduler_,
        dynamics=nh.api.Dynamics.coerce(config),
        monitor=(nh.Monitor, {
            'minimize': ['loss'],
            'patience': config['patience'],
            'max_epoch': config['max_epoch'],
            'smoothing': 0.0,
        }),
        initializer=initializer_,
        criterion=(torch.nn.CrossEntropyLoss, {}),
        # The rests of the keyword arguments are simply dictionaries used to
        # track other information.
        # Specify what augmentations you are performing for experiment tracking
        # augment=transform_train,
        other={
            # Specify anything else that is special about your hyperparams here
            # Especially if you make a custom_batch_runner
            'augment': config['augment'],
        },
        # These extra arguments are recorded in the train_info.json but do
        # not contribute to the hyperparameter hash.
        extra={
            'config': ub.repr2(dict(config)),
            'argv': sys.argv,
        }
    )

    # Creating an instance of a Fitharn object is typically fast.
    harn = CIFAR_FitHarn(hyper=hyper)
    harn.preferences['prog_backend'] = 'progiter'
    harn.preferences['keyboard_debug'] = True
    harn.preferences['eager_dump_tensorboard'] = True
    harn.preferences['tensorboard_groups'] = ['loss']
    # harn.preferences['tensorboard_groups']

    harn.intervals.update({
        'vali': 1,
        'test': 1,
    })

    harn.script_config = config
    return harn


def main():
    # Run your code that sets up your custom FitHarn object.
    harn = setup_harn()

    # Initializing a FitHarn object can take a little time, but not too much.
    # This is where instances of the model, optimizer, scheduler, monitor, and
    # initializer are created. This is also where we check if there is a
    # pre-existing checkpoint that we can restart from.
    harn.initialize()

    if ub.argflag('--lrtest'):
        """
        python -m netharn.examples.cifar --xpu=0 --arch=efficientnet-b0 \
                --nice=test_cifar9 --optim=adamw --schedule=Exponential-g0.98 \
                --lr=0.1 --init=kaiming_normal \
                --batch_size=2048 --lrtest --show

        python -m netharn.examples.cifar --xpu=0 --arch=efficientnet-b7 \
                --nice=test_cifar9 --optim=adamw --schedule=Exponential-g0.98 \
                --lr=0.1 --init=kaiming_normal \
                --batch_size=256  --lrtest --show

        python -m netharn.examples.cifar --xpu=0 --arch=efficientnet-b7 \
                --nice=test_cifar9 --optim=adamw --schedule=Exponential-g0.98 \
                --lr=4e-2 --init=kaiming_normal \
                --batch_size=256
        """
        # Undocumented hidden feature,
        # Perform an LR-test, then resetup the harness. Optionally draw the
        # results using matplotlib.
        from netharn.prefit.lr_tests import lr_range_test

        result = lr_range_test(
            harn, init_value=1e-4, final_value=0.5, beta=0.3,
            explode_factor=10, num_iters=200)

        if ub.argflag('--show'):
            import kwplot
            plt = kwplot.autoplt()
            result.draw()
            plt.show()

        # Recreate a new version of the harness with the recommended LR.
        config = harn.script_config.asdict()
        config['lr'] = (result.recommended_lr * 10)
        harn = setup_harn(**config)
        harn.initialize()

    # This starts the main loop which will run until the monitor's terminator
    # criterion is satisfied. If the initialize step loaded a checkpointed that
    # already met the termination criterion, then this will simply return.
    deploy_fpath = harn.run()

    # The returned deploy_fpath is the path to an exported netharn model.
    # This model is the on with the best weights according to the monitor.
    print('deploy_fpath = {!r}'.format(deploy_fpath))
    return harn


if __name__ == '__main__':
    r"""
    The baseline script replicates parameters from
    https://github.com/kuangliu/pytorch-cifar

    The following is a table of kuangliu's reported accuracy and our measured
    accuracy for each architecture.

    The first column is kuangliu's reported accuracy, the second column is me
    running kuangliu's code, and the final column is using my own training
    harness (handles logging and whatnot) called netharn.

    The first three experiments are with simple augmentation. The rest have
    more complex augmentation.

           arch        |  kuangliu  | rerun-kuangliu |  netharn |  train rate | num params
    ---------------------------------------------------------------------------------------
    ResNet50           |    93.62%  |        95.370% |  95.72%  |             |
    DenseNet121        |    95.04%  |        95.420% |  94.47%  |             |
    DPN92              |    95.16%  |        95.410% |  94.92%  |             |
    --------------------
    ResNet50_newaug*   |        --  |             -- |  96.13%  |   498.90 Hz | 23,520,842
    EfficientNet-7*    |        --  |             -- |  85.36%  |   214.18 Hz | 63,812,570
    EfficientNet-3*    |        --  |             -- |  86.87%  |   568.30 Hz | 10,711,602
    EfficientNet-0*    |        --  |             -- |  87.13%  |   964.21 Hz |  4,020,358

    EfficientNet-0-b64-224 |    --  |             -- |  25ish%  |   148.15 Hz |  4,020,358
    efficientnet0_transfer_b64_sz224_v2 ||           |  98.04%  |


   600025177002,


    CommandLine:
        python -m netharn.examples.cifar --xpu=0 --nice=resnet50_baseline --arch=resnet50 --optim=sgd --schedule=step-150-250 --lr=0.1
        python -m netharn.examples.cifar --xpu=0 --nice=wrn --arch=wrn_22 --optim=sgd --schedule=step-150-250 --lr=0.1
        python -m netharn.examples.cifar --xpu=0 --nice=densenet --arch=densenet121 --optim=sgd --schedule=step-150-250 --lr=0.1

        python -m netharn.examples.cifar --xpu=0 --nice=se_resnet18 --arch=se_resnet18 --optim=sgd --schedule=step-150-250 --lr=0.01 --init=noop --decay=1e-5 --augment=simple

        python -m netharn.examples.cifar --xpu=0 --nice=resnet50_newaug_b128 --batch_size=128 --arch=resnet50 --optim=sgd --schedule=step-150-250 --lr=0.1 --init=kaiming_normal --augment=simple

        python -m netharn.examples.cifar --xpu=0 --nice=efficientnet7_newaug_b128 --batch_size=128 --arch=efficientnet-b7 --optim=sgd --schedule=step-150-250 --lr=0.1 --init=kaiming_normal --augment=simple

        python -m netharn.examples.cifar --xpu=0 --nice=efficientnet3_newaug_b128 --batch_size=128 --arch=efficientnet-b3 --optim=sgd --schedule=step-150-250 --lr=0.1 --init=kaiming_normal --augment=simple

        python -m netharn.examples.cifar --xpu=0 --nice=efficientnet0_newaug_b128 --batch_size=128 --arch=efficientnet-b0 --optim=sgd --schedule=step-150-250 --lr=0.1 --init=kaiming_normal --augment=simple


        python -m netharn.examples.cifar --xpu=0 --nice=efficientnet0_transfer_b128_sz32 --batch_size=128 --arch=efficientnet-b0 --optim=sgd --schedule=step-150-250 --lr=0.01 --decay=5e-4 --init=cls --augment="crop,flip,gray,cutout" --input_dims=32,32

        python -m netharn.examples.cifar --xpu=0 --nice=efficientnet0_transfer_b64_sz224 --batch_size=64 --arch=efficientnet-b0 --optim=sgd --schedule=step-150-250 --lr=0.01 --decay=5e-4 --init=cls --augment="crop,flip,gray,cutout" --input_dims=224,224

        python -m netharn.examples.cifar --xpu=0 --nice=efficientnet0_newaug_b64_sz224 --batch_size=64 --arch=efficientnet-b0 --optim=sgd --schedule=step-150-250 --lr=0.1 --init=kaiming_normal --augment=simple --input_dims=224,224

        python -m netharn.examples.cifar --xpu=0 --nice=efficientnet0_transfer_b128_sz32_v2 --batch_size=128 --arch=efficientnet-b0 --optim=sgd --schedule=step-20-45-70-90-f5 --max_epoch=100 --lr=0.01 --decay=5e-4 --init=cls --augment="crop,flip,gray,cutout" --input_dims=32,32  # 88%

        python -m netharn.examples.cifar --xpu=0 --nice=efficientnet0_transfer_b128_sz32_v3 --batch_size=128 --arch=efficientnet-b0 --optim=sgd --schedule=step-13-20-45-70-90-f5 --max_epoch=100 --lr=0.01 --decay=5e-4 --init=cls --augment="crop,flip,gray,cutout" --input_dims=32,32

        python -m netharn.examples.cifar --xpu=0 --nice=efficientnet0_transfer_b128_sz32_v4 --batch_size=128 --arch=efficientnet-b0 --optim=sgd --schedule=step-10-20-45-70-90-f5 --max_epoch=100 --lr=0.03 --decay=5e-4 --init=cls --augment="crop,flip,gray,cutout" --input_dims=32,32

        python -m netharn.examples.cifar --xpu=0 --nice=efficientnet0_transfer_b64_sz224_v2 --batch_size=64 --arch=efficientnet-b0 --optim=sgd --schedule=step-10-20 --max_epoch=100 --lr=0.01 --decay=5e-4 --init=cls --augment="crop,flip,gray,cutout" --input_dims=224,224


        python -m netharn.examples.cifar --xpu=0 --nice=efficientnet0_newaug_yogi_b1024 \
                --batch_size=1028 --arch=efficientnet-b0 --optim=Yogi \
                --schedule=step-60-120-160-250-350-f5 --decay=5e-4 --lr=0.01549 \
                --init=kaiming_normal --augment=simple --grad_norm_max=35 \
                --warmup_iters=100

        python -m netharn.examples.cifar --xpu=0 --nice=efficientnet1_newaug_diffgrad_b1024 \
                --batch_size=1028 --arch=efficientnet-b1 --optim=DiffGrad \
                --schedule=step-60-120-160-250-350-f5 --decay=5e-4 --lr=0.01 \
                --init=kaiming_normal --augment=simple --grad_norm_max=35 \
                --warmup_iters=100


        # Params from Cutout paper: https://arxiv.org/pdf/1708.04552.pdf
        python -m netharn.examples.cifar --xpu=0 --nice=repro_cutout \
                --batch_size=128 \
                --arch=efficientnet-b0 \
                --optim=sgd --lr=0.01 --decay=5e-4 \
                --schedule=step-60-120-160-f5 --max_epoch=200 \
                --init=kaiming_normal --augment=simple \
                --grad_norm_max=35 --warmup_iters=100

        python -m netharn.examples.cifar --xpu=0 --nice=repro_cutoutDiffGrad \
                --batch_size=128 \
                --arch=efficientnet-b1 \
                --optim=DiffGrad --lr=0.01 --decay=5e-4 \
                --schedule=step-60-120-160-f5 --max_epoch=200 \
                --init=kaiming_normal --augment=simple \
                --grad_norm_max=35 --warmup_iters=100

        0.015219216761025578


        python -m netharn.examples.cifar --xpu=0 --nice=efficientnet7_scratch \
            --arch=efficientnet-b7 --optim=sgd --schedule=step-150-250-350 \
            --batch_size=512 --lr=0.01 --init=noop --decay=1e-5

    CommandLine:
        python -m netharn.examples.cifar --xpu=0 --arch=resnet50 --num_vali=0
        python -m netharn.examples.cifar --xpu=0 --arch=efficientnet-b0 --num_vali=0

        python -m netharn.examples.cifar --xpu=0 --arch=efficientnet-b0

        # This next command requires a bit more compute
        python -m netharn.examples.cifar --xpu=0 --arch=efficientnet-b0 --nice=test_cifar2 --schedule=step-3-6-50 --lr=0.1 --init=cls --batch_size=2718

        python -m netharn.examples.cifar --xpu=0 --arch=efficientnet-b0 --nice=test_cifar3 --schedule=step-3-6-12-16 --lr=0.256 --init=cls --batch_size=3000 --workers=2 --num_vali=0 --optim=rmsprop

        python -m netharn.examples.cifar --xpu=0 --arch=efficientnet-b0 --nice=test_cifar3 --schedule=onecycle70 --lr=0.01  --init=cls --batch_size=3000 --workers=2 --num_vali=0 --optim=sgd --datasets=cifar100

        python -m netharn.examples.cifar --xpu=0 --arch=efficientnet-b0 --nice=test_cifar2 --schedule=ReduceLROnPlateau-p1-c1-f0.9 --lr=0.1 --init=cls --batch_size=2719 --workers=4 --optim=sgd --datasets=cifar100

        python -m netharn.examples.cifar.py --xpu=0 --arch=densenet121
        # Train on two GPUs with a larger batch size
        python -m netharn.examples.cifar.py --arch=dpn92 --batch_size=256 --xpu=0,1
    """
    import seaborn
    seaborn.set()
    main()



"""
Modified from:
    https://github.com/ducha-aiki/LSUV-pytorch
"""
import numpy as np
import torch
import torch.nn.init
import torch.nn as nn
import ubelt as ub
from netharn import util
from netharn import api
from netharn.initializers.functional import trainable_layers


def svd_orthonormal(shape, rng=None, cache_key=None):
    """
    If cache_key is specified, then the result will be cached, and subsequent
    calls with the same key and shape will return the same result.

    References:
        Orthonorm init code is taked from Lasagne
        https://github.com/Lasagne/Lasagne/blob/master/lasagne/init.py
    """
    rng = util.ensure_rng(rng)

    if len(shape) < 2:
        raise RuntimeError("Only shapes of length 2 or more are supported.")
    flat_shape = (shape[0], np.prod(shape[1:]))

    enabled = False and cache_key is not None
    if enabled:
        rand_sequence = rng.randint(0, 2 ** 16)
        depends = [shape, cache_key, rand_sequence]
        cfgstr = ub.hash_data(depends)
    else:
        cfgstr = ''

    # this process can be expensive, cache it

    # TODO: only cache very large matrices (4096x4096)
    # TODO: only cache very large matrices, not (256,256,3,3)
    cacher = ub.Cacher('svd_orthonormal', appname='netharn', enabled=enabled,
                       cfgstr=cfgstr)
    q = cacher.tryload()
    if q is None:
        # print('Compute orthonormal matrix with shape ' + str(shape))
        a = rng.normal(0.0, 1.0, flat_shape)
        u, _, v = np.linalg.svd(a, full_matrices=False)
        q = u if u.shape == flat_shape else v
        # print(shape, flat_shape)
        q = q.reshape(shape)
        q = q.astype(np.float32)
        cacher.save(q)
    return q


class Orthonormal(api.Initializer):
    def __init__(self, rng=None):
        self.rng = util.ensure_rng(rng)

    def forward(self, model):
        for name, m in trainable_layers(model, names=True):
            pass
            if isinstance(m, torch.nn.modules.conv._ConvNd) or isinstance(m, nn.Linear):
                if hasattr(m, 'weight_v'):
                    shape = tuple(m.weight_v.shape)
                    w_ortho = svd_orthonormal(shape, self.rng, cache_key=name)
                    m.weight_v.data[:] = torch.from_numpy(w_ortho)
                    try:
                        nn.init.constant_(m.bias, 0)
                    except Exception:
                        pass
                else:
                    shape = tuple(m.weight.shape)
                    w_ortho = svd_orthonormal(shape, self.rng, cache_key=name)
                    m.weight.data[:] = torch.from_numpy(w_ortho)
                    try:
                        nn.init.constant_(m.bias, 0)
                    except Exception:
                        pass
        return model


class LSUV(api.Initializer):
    """
    CommandLine:
        python -m netharn.initializers.lsuv LSUV:0

    Example:
        >>> # xdoc: +REQUIRES(--slow)
        >>> from netharn.initializers.lsuv import *
        >>> import torchvision
        >>> import torch
        >>> #model = torchvision.models.AlexNet()
        >>> model = torchvision.models.SqueezeNet()
        >>> initer = LSUV(rng=0)
        >>> data = torch.autograd.Variable(torch.randn(4, 3, 224, 224))
        >>> initer.forward(model, data)

    Example:
        >>> # DISABLE_DOCTEST
        >>> from netharn.initializers.lsuv import *
        >>> import torchvision
        >>> import torch
        >>> model = torchvision.models.AlexNet()
        >>> initer = LSUV(rng=0)
        >>> data = torch.autograd.Variable(torch.randn(4, 3, 224, 224))
        >>> initer.forward(model, data)

    Example:
        >>> # DISABLE_DOCTEST
        >>> from netharn.initializers.lsuv import *
        >>> import torchvision
        >>> import torch
        >>> model = torchvision.models.DenseNet()
        >>> initer = LSUV(rng=0)
        >>> data = torch.autograd.Variable(torch.randn(4, 3, 224, 224))
        >>> initer.forward(model, data)
    """
    def __init__(self, needed_std=1.0, std_tol=0.1, max_attempts=10,
                 do_orthonorm=True, rng=None):

        self.rng = util.ensure_rng(rng)

        self.do_orthonorm = do_orthonorm
        self.needed_std = needed_std
        self.std_tol = std_tol
        self.max_attempts = max_attempts

    def apply_weights_correction(self, m):
        if self.gg['hook'] is None:
            return
        if not self.gg['correction_needed']:
            return
        if (isinstance(m, torch.nn.modules.conv._ConvNd)) or (isinstance(m, nn.Linear)):
            if self.gg['counter_to_apply_correction'] < self.gg['hook_position']:
                self.gg['counter_to_apply_correction'] += 1
            else:
                if hasattr(m, 'weight_g'):
                    m.weight_g.data *= float(self.gg['current_coef'])
                    #print(m.weight_g.data)
                    #print(m.weight_v.data)
                    #print('weights norm after = ', m.weight.data.norm())
                    self.gg['correction_needed'] = False
                else:
                    #print('weights norm before = ', m.weight.data.norm())
                    m.weight.data *= self.gg['current_coef']
                    #print('weights norm after = ', m.weight.data.norm())
                    self.gg['correction_needed'] = False
                return
        return

    def store_activations(self, module, input, output):
        self.gg['act_dict'] = output.data.cpu().numpy()
        return

    def add_current_hook(self, m):
        if self.gg['hook'] is not None:
            return
        if (isinstance(m, torch.nn.modules.conv._ConvNd)) or (isinstance(m, nn.Linear)):
            #print('trying to hook to', m, self.gg['hook_position'], self.gg['done_counter'])
            if self.gg['hook_position'] > self.gg['done_counter']:
                self.gg['hook'] = m.register_forward_hook(self.store_activations)
                #print(' hooking layer = ', self.gg['hook_position'], m)
            else:
                #print(m, 'already done, skipping')
                self.gg['hook_position'] += 1
        return

    def count_conv_fc_layers(self, m):
        if (isinstance(m, torch.nn.modules.conv._ConvNd)) or (isinstance(m, nn.Linear)):
            self.gg['total_fc_conv_layers'] += 1
        return

    def remove_hooks(self, hooks):
        for h in hooks:
            h.remove()
        return

    def forward(self, model, data):
        import tqdm
        self.gg = {}
        self.gg['hook_position'] = 0
        self.gg['total_fc_conv_layers'] = 0
        self.gg['done_counter'] = -1
        self.gg['hook'] = None
        self.gg['act_dict'] = {}
        self.gg['counter_to_apply_correction'] = 0
        self.gg['correction_needed'] = False
        self.gg['current_coef'] = 1.0

        model.train(False)

        print('Starting LSUV')
        model.apply(self.count_conv_fc_layers)

        print('Total layers to process:', self.gg['total_fc_conv_layers'])
        if self.do_orthonorm:
            print('Applying orthogonal weights')
            Orthonormal(rng=self.rng).forward(model)
            # model.apply(self.orthogonal_weights_init)
            print('Orthonorm done')
            # if cuda:
            #     model = model.cuda()

        for layer_idx in tqdm.trange(self.gg['total_fc_conv_layers'], desc='init layer', leave=True):
            # print(layer_idx)
            model.apply(self.add_current_hook)
            out = model(data)  # NOQA
            current_std = self.gg['act_dict'].std()
            # tqdm.tqdm.write('layer {}: std={:.4f}'.format(layer_idx, current_std))
            #print  self.gg['act_dict'].shape
            attempts = 0
            for attempts in range(self.max_attempts):
                if not (np.abs(current_std - self.needed_std) > self.std_tol):
                    break
                self.gg['current_coef'] =  self.needed_std / (current_std  + 1e-8)
                self.gg['correction_needed'] = True
                model.apply(self.apply_weights_correction)

                # if cuda:
                #     model = model.cuda()

                out = model(data)  # NOQA

                current_std = self.gg['act_dict'].std()
                # tqdm.tqdm.write('layer {}: std={:.4f}, mean={:.4f}'.format(
                #         layer_idx, current_std, self.gg['act_dict'].mean()))
            if attempts + 1 >= self.max_attempts:
                tqdm.tqdm.write('Cannot converge in {} iterations'.format(self.max_attempts))
            if self.gg['hook'] is not None:
                self.gg['hook'].remove()
            self.gg['done_counter'] += 1
            self.gg['counter_to_apply_correction'] = 0
            self.gg['hook_position'] = 0
            self.gg['hook']  = None
            # print('finish at layer', layer_idx)
        print('LSUV init done!')

        # if not cuda:
        #     model = model.cpu()
        return model

if __name__ == '__main__':
    r"""
    CommandLine:
        python -m netharn.initializers.lsuv
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
from collections import OrderedDict
import imgaug


class ParamatarizedAugmenter(imgaug.augmenters.Augmenter):
    """
    Helper that automatically registers stochastic parameters
    """

    def __init__(self, *args, **kwargs):
        super(ParamatarizedAugmenter, self).__setattr__('_initialized', True)
        super(ParamatarizedAugmenter, self).__setattr__('_registered_params', OrderedDict())
        super(ParamatarizedAugmenter, self).__init__(*args, **kwargs)

    def _setparam(self, name, value):
        self._registered_params[name] = value
        setattr(self, name, value)

    def get_parameters(self):
        return list(self._registered_params.values())

    def __setattr__(self, key, value):
        if not getattr(self, '_initialized', False) and key != '_initialized':
            raise Exception(
                ('Must call super().__init__ in {} that inherits '
                 'from Augmenter2').format(self.__class__))
        if not key.startswith('_'):
            if key in self._registered_params:
                self._registered_params[key] = value
            elif isinstance(value, imgaug.parameters.StochasticParameter):
                self._registered_params[key] = value
        super(ParamatarizedAugmenter, self).__setattr__(key, value)

    def _augment_heatmaps(self):
        raise NotImplementedError

    @staticmethod
    def _hack_get_named_params(self):
        """ hopefully imgaug will add get_named_params. Until then hack it. """
        named_params = OrderedDict()
        params = self.get_parameters()
        if params:
            # See if we can hack to gether what the param names are
            unused = OrderedDict(sorted(self.__dict__.items()))
            for p in params:
                found = False
                for key, value in list(unused.items()):
                    if p is value:
                        named_params[key] = p
                        unused.pop(key)
                        found = True
                if not found:
                    key = '__UNKNOWN_PARAM_NAME_{}__'.format(len(named_params))
                    named_params[key] = p
        return named_params

    @staticmethod
    def _json_id(aug):
        """
        TODO:
            - [ ] submit a PR to imgaug that registers parameters with classes

        Example:
            >>> from netharn.data.transforms.augmenter_base import *
            >>> import imgaug.augmenters as iaa
            >>> import imgaug
            >>> _PA = ParamatarizedAugmenter
            >>> augment = imgaug.augmenters.Affine()
            >>> info = _PA._json_id(augment)
            >>> assert info['__class__'] == 'Affine'
            >>> assert _PA._json_id('') == ''
            >>> #####
            >>> augmentors = [
            >>>     iaa.Fliplr(p=.5),
            >>>     iaa.Flipud(p=.5),
            >>>     iaa.Affine(
            >>>         scale={"x": (1.0, 1.01), "y": (1.0, 1.01)},
            >>>         translate_percent={"x": (-0.1, 0.1), "y": (-0.1, 0.1)},
            >>>         rotate=(-15, 15),
            >>>         shear=(-7, 7),
            >>>         order=[0, 1, 3],
            >>>         cval=(0, 255),
            >>>         mode=imgaug.ALL,  # use any of scikit-image's warping modes (see 2nd image from the top for examples)
            >>>         # Note: currently requires imgaug master version
            >>>         backend='cv2',
            >>>     ),
            >>>     iaa.AddToHueAndSaturation((-20, 20)),  # change hue and saturation
            >>>     iaa.ContrastNormalization((0.5, 2.0), per_channel=0.5),  # improve or worsen the contrast
            >>> ]
            >>> augment = iaa.Sequential(augmentors)
            >>> info = _PA._json_id(augment)
            >>> import ubelt as ub
            >>> print(ub.repr2(info, nl=2, precision=2))
        """
        _PA = ParamatarizedAugmenter
        if isinstance(aug, tuple):
            return [_PA._json_id(item) for item in aug]
        elif isinstance(aug, imgaug.parameters.StochasticParameter):
            return str(aug)
        elif isinstance(aug, imgaug.augmenters.Augmenter):
            info = OrderedDict()
            info['__class__'] = aug.__class__.__name__
            try:
                params = _PA._hack_get_named_params(aug)
                if params:
                    info['params'] = params
                if isinstance(aug, list):
                    children = aug[:]
                    children = [ParamatarizedAugmenter._json_id(c) for c in children]
                    info['children'] = children
                return info
            except Exception as ex:
                print(ex)
                # imgaug is weird and buggy
                info['__str__'] = str(aug)
        else:
            return str(aug)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
from os.path import exists
import io
from os.path import join
from os.path import os
import tempfile
import zipfile
import ubelt as ub
import re


def split_archive(fpath):
    """
    If fpath specifies a file inside a zipfile, it breaks it into two parts the
    path to the zipfile and the internal path in the zipfile.

    fpath = '/'
    split_archive('/a/b/foo.zip/bar.txt')
    split_archive('/a/b/foo.zip/baz/bar.txt')
    split_archive('/a/b/foo.zip/baz/biz.zip/bar.txt')
    split_archive('/a/b/foo.zip/baz/biz.zip/bar.py')
    """
    pat = '(.zip[' + re.escape(os.path.sep) + '/:])'
    parts = re.split(pat, fpath, flags=re.IGNORECASE)
    if len(parts) > 2:
        archivepath = ''.join(parts[:-1])[:-1]
        internal = parts[-1]
    else:
        archivepath = None
        internal = None
    return archivepath, internal


class zopen(ub.NiceRepr):
    """
    Can open a file normally or open a file within a zip file (readonly). Tries
    to read from memory only, but will extract to a tempfile if necessary.

    Example:
        >>> import torch
        >>> dpath = ub.ensure_app_cache_dir('netharn')
        >>> data = torch.FloatTensor([0])
        >>> torch.save(data, join(dpath, 'data.pt'))
        >>> torch.load(join(dpath, 'data.pt'))
        >>> torch.load(open(join(dpath, 'data.pt'), 'rb'))
        >>> torch.load(zopen(join(dpath, 'data.pt'), 'rb'))

    Example:
        >>> import torch
        >>> # Test we can load torch data from a zipfile
        >>> dpath = ub.ensure_app_cache_dir('netharn')
        >>> data = torch.FloatTensor([0])
        >>> datapath = join(dpath, 'data.pt')
        >>> torch.save(data, datapath)
        >>> zippath = join(dpath, 'datazip.zip')
        >>> internal = 'folder/data.pt'
        >>> with zipfile.ZipFile(zippath, 'w') as myzip:
        >>>     myzip.write(datapath, internal)
        >>> fpath = zippath + '/' + internal
        >>> file = zopen(fpath, 'rb', seekable=True)
        >>> data2 = torch.load(file._handle)
        >>> file = zopen(datapath, 'rb', seekable=True)
        >>> data3 = torch.load(file._handle)

    Example:
        >>> # Test we can load json data from a zipfile
        >>> dpath = ub.ensure_app_cache_dir('netharn')
        >>> infopath = join(dpath, 'info.json')
        >>> open(infopath, 'w').write('{"x": "1"}')
        >>> zippath = join(dpath, 'infozip.zip')
        >>> internal = 'folder/info.json'
        >>> with zipfile.ZipFile(zippath, 'w') as myzip:
        >>>     myzip.write(infopath, internal)
        >>> fpath = zippath + '/' + internal
        >>> file = zopen(fpath, 'r')
        >>> import json
        >>> info2 = json.load(file)
        >>> assert info2['x'] == '1'
    """
    def __init__(self, fpath, mode='r', seekable=False):
        self.fpath = fpath
        self.name = fpath
        self.mode = mode
        self._seekable = seekable
        assert 'r' in self.mode
        self._handle = None
        self._zfpath = None
        self._temp_dpath = None
        self._temp_fpath = None
        self._open()

    def __nice__(self):
        if self._zfpath is None:
            return str(self._handle) + ' mode=' + self.mode
        else:
            return '{} in zipfile {}, mode={}'.format(self._handle, self._zfpath, self.mode)

    def __getattr__(self, key):
        # Expose attributes of wrapped handle
        if hasattr(self._handle, key):
            return getattr(self._handle, key)
        raise AttributeError(key)

    def __dir__(self):
        # Expose attributes of wrapped handle
        keyset = set(dir(super(zopen, self)))
        keyset.update(set(self.__dict__.keys()))
        if self._handle is not None:
            keyset.update(set(dir(self._handle)))
        return sorted(keyset)

    def _cleanup(self):
        # print('self._cleanup = {!r}'.format(self._cleanup))
        if not getattr(self, 'closed', True):
            getattr(self, 'close', lambda: None)()
        if self._temp_dpath and exists(self._temp_dpath):
            ub.delete(self._temp_dpath)

    def __del__(self):
        self._cleanup()

    def _open(self):
        _handle = None
        if exists(self.fpath):
            _handle = open(self.fpath, self.mode)
        elif '.zip/' in self.fpath or '.zip' + os.path.sep in self.fpath:
            fpath = self.fpath
            archivefile, internal = split_archive(fpath)
            myzip = zipfile.ZipFile(archivefile, 'r')
            if self._seekable:
                # If we need data to be seekable, then we must extract it to a
                # temporary file first.
                self._temp_dpath = tempfile.mkdtemp()
                temp_fpath = join(self._temp_dpath, internal)
                myzip.extract(internal, self._temp_dpath)
                _handle = open(temp_fpath, self.mode)
            else:
                # Try to load data directly from the zipfile
                _handle = myzip.open(internal, 'r')
                if self.mode == 'rb':
                    data = _handle.read()
                    _handle = io.BytesIO(data)
                elif self.mode == 'r':
                    # FIXME: doesnt always work. handle seems to be closed too
                    # soon in the case util.zopen(module.__file__).read()
                    _handle = io.TextIOWrapper(_handle)
                else:
                    raise KeyError(self.mode)
                self._zfpath = archivefile
        if _handle is None:
            raise IOError('file {!r} does not exist'.format(self.fpath))
        self._handle = _handle

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m netharn.util.util_zip all
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



import os
from collections import OrderedDict
import sys
import datetime


def get_file_info(fpath, raw=False):
    """
    Get OS-level info about a file path.

    Args:
        fpath (PathLike): a path to a file.
        raw (bool): if True returns raw counter timestamps, otherwise
            returns datetime wrapper objects.

    Returns:
        Dict: containing
            created: Time of creation (varies by OS, see os.stat docs)
            filesize: Size in bytes of the file
            last_accessed: Time of last access.
            last_modified: Time of last modification.
            owner: user that owns the file (None if unable to be determined)

    Example:
        >>> import ubelt as ub
        >>> fpath = ub.__file__
        >>> info = get_file_info(fpath)
        >>> print(ub.repr2(info, nl=1))  # xdoc: +IGNORE_WANT
        {
            'created': datetime.datetime(2018, 10, 25, 12, 30, 36, 806656),
            'filesize': 5868,
            'last_accessed': datetime.datetime(2018, 10, 25, 12, 30, 38, 886705),
            'last_modified': datetime.datetime(2018, 10, 25, 12, 30, 36, 806656),
            'owner': 'joncrall',
        }
    """
    statbuf = os.stat(fpath)

    try:
        # Sometimes this fails
        if sys.platform.startswith('win32'):
            import win32security
            sec_desc = win32security.GetFileSecurity(
                fpath, win32security.OWNER_SECURITY_INFORMATION)
            owner_sid = sec_desc.GetSecurityDescriptorOwner()
            owner = win32security.LookupAccountSid(None, owner_sid)[0]
        else:
            from pwd import getpwuid
            owner = getpwuid(statbuf.st_uid).pw_name
    except Exception:
        owner = None

    info = OrderedDict([
        ('created', statbuf.st_ctime),
        ('filesize', statbuf.st_size),
        ('last_accessed', statbuf.st_atime),
        ('last_modified', statbuf.st_mtime),
        ('owner', owner)
    ])
    # permission = [os.access(fpath, os.W_OK), os.access(fpath, os.X_OK)]

    if not raw:
        time_keys = [
            'last_accessed',
            'last_modified',
            'created',
        ]
        for key in time_keys:
            info[key] = datetime.datetime.fromtimestamp(info[key])
    return info



from torch import nn
import torch
import torch.nn.functional as F


@torch.jit.script
def mish(input):
    """
    Applies the mish function element-wise:
    mish(x) = x * tanh(softplus(x)) = x * tanh(ln(1 + exp(x)))
    See additional documentation for mish class.
    """
    return input * torch.tanh(F.softplus(input))


def beta_mish(input, beta=1.5):
    """
    Applies the β mish function element-wise:
        .. math::
            \\beta mish(x) = x * tanh(ln((1 + e^{x})^{\\beta}))
    See additional documentation for :mod:`echoAI.Activation.Torch.beta_mish`.

    References:
        https://github.com/digantamisra98/Echo/blob/master/echoAI/Activation/Torch/functional.py
    """
    return input * torch.tanh(torch.log(torch.pow((1 + torch.exp(input)), beta)))


class Mish_Function(torch.autograd.Function):

    """
    Applies the mish function element-wise:
    .. math::
        mish(x) = x * tanh(softplus(x)) = x * tanh(ln(1 + e^{x}))
    Plot:
    .. figure::  _static/mish.png
        :align:   center
    Shape:
        - Input: (N, *) where * means, any number of additional
          dimensions
        - Output: (N, *), same shape as the input

    References:
        https://github.com/digantamisra98/Echo/blob/master/echoAI/Activation/Torch/mish.py

    Examples:
        >>> m = Mish()
        >>> input = torch.randn(2)
        >>> output = m(input)
    """

    @staticmethod
    def forward(ctx, x):
        ctx.save_for_backward(x)
        y = x * torch.tanh(F.softplus(x))  # x * tanh(ln(1 + exp(x)))
        return y

    @staticmethod
    def backward(ctx, grad_output):
        x = ctx.saved_variables[0]
        sigmoid = torch.sigmoid(x)
        tanh_sp = torch.tanh(F.softplus(x))
        return grad_output * (tanh_sp + x * sigmoid * (1 - tanh_sp * tanh_sp))

    # else:
    #     @torch.jit.script
    #     def mish(input):
    #         delta = torch.exp(-input)
    #         alpha = 1 + 2 * delta
    #         return input * alpha / (alpha + 2 * delta * delta)


class Mish(nn.Module):
    """
    Applies the mish function element-wise:
    mish(x) = x * tanh(softplus(x)) = x * tanh(ln(1 + exp(x)))

    Shape:
        - Input: (N, *) where * means, any number of additional
          dimensions
        - Output: (N, *), same shape as the input

    References:
        https://github.com/digantamisra98/Mish/blob/master/Mish/Torch/mish.py
        https://github.com/thomasbrandon/mish-cuda
        https://arxiv.org/pdf/1908.08681v2.pdf

    Examples:
        >>> m = Mish()
        >>> input = torch.randn(2)
        >>> output = m(input)

    Example:
        >>> x = torch.linspace(-20, 20, 100, requires_grad=True)
        >>> self = Mish()
        >>> y = self(x)
        >>> y.sum().backward()
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.multi_plot(xydata={'beta=1': (x.data, y.data)}, fnum=1, pnum=(1, 2, 1))
        >>> kwplot.multi_plot(xydata={'beta=1': (x.data, x.grad)}, fnum=1, pnum=(1, 2, 2))
        >>> kwplot.show_if_requested()
    """
    def __init__(self):
        super().__init__()

    def forward(self, input):
        return Mish_Function.apply(input)
        # return mish(input)



import six
import torch
import ubelt as ub
if six.PY2:
    from fractions import gcd
else:
    from math import gcd


def rectify_nonlinearity(key=ub.NoParam, dim=2):
    """
    Allows dictionary based specification of a nonlinearity

    Example:
        >>> rectify_nonlinearity('relu')
        ReLU(...)
        >>> rectify_nonlinearity('leaky_relu')
        LeakyReLU(negative_slope=0.01...)
        >>> rectify_nonlinearity(None)
        None
        >>> rectify_nonlinearity('swish')
    """
    if key is None:
        return None

    if key is ub.NoParam:
        key = 'relu'

    if isinstance(key, six.string_types):
        key = {'type': key}
    elif isinstance(key, dict):
        key = key.copy()
    else:
        raise TypeError(type(key))
    kw = key
    noli_type = kw.pop('type')
    if 'inplace' not in kw:
        kw['inplace'] = True

    if noli_type == 'leaky_relu':
        cls = torch.nn.LeakyReLU
    elif noli_type == 'relu':
        cls = torch.nn.ReLU
    elif noli_type == 'elu':
        cls = torch.nn.ELU
    elif noli_type == 'celu':
        cls = torch.nn.CELU
    elif noli_type == 'selu':
        cls = torch.nn.SELU
    elif noli_type == 'relu6':
        cls = torch.nn.ReLU6
    elif noli_type == 'swish':
        from netharn.layers.swish import Swish
        kw.pop('inplace', None)
        cls = Swish
    elif noli_type == 'mish':
        from netharn.layers.mish import Mish
        kw.pop('inplace', None)
        cls = Mish
    else:
        raise KeyError('unknown type: {}'.format(kw))
    return cls(**kw)


def rectify_normalizer(in_channels, key=ub.NoParam, dim=2, **kwargs):
    """
    Allows dictionary based specification of a normalizing layer

    Args:
        in_channels (int): number of input channels
        dim (int): dimensionality
        **kwargs: extra args

    Example:
        >>> rectify_normalizer(8)
        BatchNorm2d(8, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
        >>> rectify_normalizer(8, 'batch')
        BatchNorm2d(8, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
        >>> rectify_normalizer(8, {'type': 'batch'})
        BatchNorm2d(8, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
        >>> rectify_normalizer(8, 'group')
        GroupNorm(8, 8, eps=1e-05, affine=True)
        >>> rectify_normalizer(8, {'type': 'group', 'num_groups': 2})
        GroupNorm(2, 8, eps=1e-05, affine=True)
        >>> rectify_normalizer(8, dim=3)
        BatchNorm3d(8, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
        >>> rectify_normalizer(8, None)
        None
        >>> rectify_normalizer(8, key={'type': 'syncbatch'})
    """
    if key is None:
        return None

    if key is ub.NoParam:
        key = 'batch'

    if isinstance(key, six.string_types):
        key = {'type': key}
    elif isinstance(key, dict):
        key = key.copy()
    else:
        raise TypeError(type(key))

    norm_type = key.pop('type')
    if norm_type == 'batch':
        in_channels_key = 'num_features'

        if dim == 0:
            cls = torch.nn.BatchNorm1d
        elif dim == 1:
            cls = torch.nn.BatchNorm1d
        elif dim == 2:
            cls = torch.nn.BatchNorm2d
        elif dim == 3:
            cls = torch.nn.BatchNorm3d
        else:
            raise ValueError(dim)
    elif norm_type == 'syncbatch':
        in_channels_key = 'num_features'
        cls = torch.nn.SyncBatchNorm
    elif norm_type == 'group':
        in_channels_key = 'num_channels'
        if key.get('num_groups') is None:
            key['num_groups'] = ('gcd', min(in_channels, 32))

        if isinstance(key['num_groups'], tuple):
            if key['num_groups'][0] == 'gcd':
                key['num_groups'] = gcd(
                    key['num_groups'][1], in_channels)
        if in_channels % key['num_groups'] != 0:
            raise AssertionError(
                'Cannot divide n_inputs {} by num groups {}'.format(
                    in_channels, key['num_groups']))
        cls = torch.nn.GroupNorm
    elif norm_type == 'batch+group':
        return torch.nn.Sequential(
            rectify_normalizer(in_channels, 'batch', dim=dim),
            rectify_normalizer(in_channels, ub.dict_union({'type': 'group'}, key), dim=dim),
        )
    else:
        raise KeyError('unknown type: {}'.format(key))
    assert in_channels_key not in key
    key[in_channels_key] = in_channels

    try:
        import copy
        kw = copy.copy(key)
        kw.update(kwargs)
        return cls(**kw)
    except Exception:
        # Ignore kwargs
        import warnings
        warnings.warn('kwargs ignored in rectify normalizer')
        return cls(**key)


def rectify_conv(dim=2):
    conv_cls = {
        0: torch.nn.Linear,
        1: torch.nn.Conv1d,
        2: torch.nn.Conv2d,
        3: torch.nn.Conv3d,
    }[dim]
    return conv_cls


def rectify_dropout(dim=2):
    conv_cls = {
        0: torch.nn.Dropout,
        1: torch.nn.Dropout,
        2: torch.nn.Dropout2d,
        3: torch.nn.Dropout3d,
    }[dim]
    return conv_cls


def rectify_maxpool(dim=2):
    conv_cls = {
        0: torch.nn.MaxPool1d,
        1: torch.nn.MaxPool1d,
        2: torch.nn.MaxPool2d,
        3: torch.nn.MaxPool3d,
    }[dim]
    return conv_cls



"""
Under development! Function names and logic may change at any time. Nothing in
this file should be considered as stable! Use at your own risk.

These are methods that you can mixin to your FitHarn implementation to extend
its functionality to typical, but non-default cases.

The purpose of this file is to contain functions that might not general-purpose
enough to add to FitHarn itself, but they are also common enough, where it
makes no sense to write them from scratch for each new project.
"""


def _dump_monitor_tensorboard(harn, mode='epoch', special_groupers=['loss'],
                              serial=False):
    """
    Dumps PNGs to disk visualizing tensorboard scalars.
    Also dumps pickles to disk containing the same information.

    Args:
        mode : can be either epoch or iter

    CommandLine:
        xdoctest -m netharn.mixins _dump_monitor_tensorboard --profile

    Example:
        >>> from netharn.export.deployer import _demodata_toy_harn
        >>> from netharn.mixins import _dump_monitor_tensorboard
        >>> harn = _demodata_toy_harn()
        >>> harn.run()
        >>> try:
        >>>     _dump_monitor_tensorboard(harn)
        >>> except ImportError:
        >>>     pass
    """
    import ubelt as ub
    import netharn as nh
    from os.path import join
    import json
    import six
    from six.moves import cPickle as pickle

    harn.debug('Plotting tensorboard data. serial={}, mode={}'.format(serial, mode))

    train_dpath = harn.train_dpath

    tb_data = nh.util.read_tensorboard_scalars(train_dpath, cache=0, verbose=0)

    tb_data['meta'] = {
        'nice': harn.hyper.nice,
        'special_groupers': special_groupers,
    }

    out_dpath = ub.ensuredir((train_dpath, 'monitor', 'tensorboard'))

    tb_data_pickle_fpath = join(out_dpath, 'tb_data.pkl')
    with open(tb_data_pickle_fpath, 'wb') as file:
        pickle.dump(tb_data, file)

    tb_data_json_fpath = join(out_dpath, 'tb_data.json')
    with open(tb_data_json_fpath, 'w') as file:
        if six.PY2:
            jsonkw = dict(indent=1)
        else:
            jsonkw = dict(indent=' ')
        try:
            json.dump(tb_data, file, **jsonkw)
        except Exception as ex:
            print('ex = {!r}'.format(ex))
            json.dump({
                'error': 'Unable to write to json.',
                'info': 'See pickle file: {}'.format(tb_data_json_fpath)},
                file, **jsonkw)

    # The following function draws the tensorboard result
    # This might take a some non-trivial amount of time so we attempt to run in
    # a separate process.
    func = _dump_measures
    args = (tb_data, out_dpath, mode)

    if not serial:

        if False:
            # Maybe thread-safer way of doing this? Maybe not, there is a
            # management thread used by futures.
            from concurrent import futures
            if not hasattr(harn, '_internal_executor'):
                harn._internal_executor = futures.ProcessPoolExecutor(max_workers=1)
                harn._prev_job = None
            if harn._prev_job is None or harn._prev_job.done():
                # Wait to before submitting another job
                # Unsure if its ok that this job might not be a daemon
                harn.info('DO MPL DRAW')
                job = harn._internal_executor.submit(func, *args)
                harn._prev_job = job
            else:
                if harn._prev_job is not None:
                    harn.info('NOT DOING MPL DRAW')
                    harn.warn('NOT DOING MPL DRAW')
        else:
            # This causes thread-unsafe warning messages in the inner loop
            # Likely because we are forking while a thread is alive
            if not hasattr(harn, '_internal_procs'):
                harn._internal_procs = ub.ddict(dict)

            # Clear finished processes from the pool
            for pid in list(harn._internal_procs[mode].keys()):
                proc = harn._internal_procs[mode][pid]
                if not proc.is_alive():
                    harn._internal_procs[mode].pop(pid)

            # only start a new process if there is room in the pool
            if len(harn._internal_procs[mode]) < 1:
                import multiprocessing
                proc = multiprocessing.Process(target=func, args=args)
                proc.daemon = True
                proc.start()
                harn._internal_procs[mode][proc.pid] = proc
            else:
                if 0:
                    harn.warn('NOT DOING MPL DRAW')
    else:
        func(*args)


def _redump_measures(dpath):
    """
    """
    import json
    from os.path import join

    import kwplot
    kwplot.set_mpl_backend('agg')

    try:
        import seaborn as sns
        sns.set()
    except ImportError:
        pass

    fpath = join(dpath, 'tb_data.json')
    tb_data = json.load(open(fpath, 'r'))

    out_dpath = dpath
    mode = 'epoch'
    _dump_measures(tb_data, out_dpath, mode)


def _dump_measures(tb_data, out_dpath, mode=None, smoothing=0.0,
                   ignore_outliers=True):
    """
    This is its own function in case we need to modify formatting

    CommandLine:
        xdoctest -m netharn.mixins _dump_measures --out_dpath=.

    Example:
        >>> # SCRIPT
        >>> # Reread a dumped pickle file
        >>> from netharn.mixins import *  # NOQA
        >>> from netharn.mixins import _dump_monitor_tensorboard, _dump_measures
        >>> import json
        >>> from os.path import join
        >>> import ubelt as ub
        >>> try:
        >>>     import seaborn as sns
        >>>     sns.set()
        >>> except ImportError:
        >>>     pass
        >>> out_dpath = ub.expandpath('~/work/project/fit/nice/nicename/monitor/tensorboard/')
        >>> out_dpath = ub.argval('--out_dpath', default=out_dpath)
        >>> mode = ['epoch', 'iter']
        >>> fpath = join(out_dpath, 'tb_data.json')
        >>> tb_data = json.load(open(fpath, 'r'))
        >>> _dump_measures(tb_data,  out_dpath, smoothing=0)
    """
    import ubelt as ub
    from os.path import join
    import numpy as np

    import kwplot
    kwplot.autompl()

    # TODO: Is it possible to get htop to show this process with some name that
    # distinguishes it from the dataloader workers?
    # import sys
    # import multiprocessing
    # if multiprocessing.current_process().name != 'MainProcess':
    #     if sys.platform.startswith('linux'):
    #         import ctypes
    #         libc = ctypes.cdll.LoadLibrary('libc.so.6')
    #         title = 'Netharn MPL Dump Measures'
    #         libc.prctl(len(title), title, 0, 0, 0)

    # NOTE: This cause warnings when exeucted as daemon process
    # try:
    #     import seaborn as sbn
    #     sbn.set()
    # except ImportError:
    #     pass

    valid_modes = ['epoch', 'iter']
    if mode is None:
        mode = valid_modes
    if ub.iterable(mode):
        # Hack: Call with all modes
        for mode_ in mode:
            _dump_measures(tb_data, out_dpath, mode=mode_, smoothing=smoothing,
                           ignore_outliers=ignore_outliers)
        return
    else:
        assert mode in valid_modes

    meta = tb_data.get('meta', {})
    nice = meta.get('nice', '?nice?')
    special_groupers = meta.get('special_groupers', ['loss'])

    fig = kwplot.figure(fnum=1)

    plot_keys = [key for key in tb_data if
                 ('train_' + mode in key or
                  'vali_' + mode in key or
                  'test_' + mode in key or
                  mode + '_' in key)]
    y01_measures = ['_acc', '_ap', '_mAP', '_auc', '_mcc', '_brier', '_mauc']
    y0_measures = ['error', 'loss']

    keys = set(tb_data.keys()).intersection(set(plot_keys))

    # print('mode = {!r}'.format(mode))
    # print('tb_data.keys() = {!r}'.format(tb_data.keys()))
    # print('plot_keys = {!r}'.format(plot_keys))
    # print('keys = {!r}'.format(keys))

    def smooth_curve(ydata, beta):
        """
        Curve smoothing algorithm used by tensorboard
        """
        import pandas as pd
        alpha = 1.0 - beta
        if alpha <= 0:
            return ydata
        ydata_smooth = pd.Series(ydata).ewm(alpha=alpha).mean().values
        return ydata_smooth

    def inlier_ylim(ydatas):
        """
        outlier removal used by tensorboard
        """
        low, high = None, None
        for ydata in ydatas:
            q1 = 0.05
            q2 = 0.95
            low_, high_ = np.quantile(ydata, [q1, q2])

            # Extrapolate how big the entire span should be based on inliers
            inner_q = q2 - q1
            inner_extent = high_ - low_
            extrap_total_extent = inner_extent  / inner_q

            # amount of padding to add to either side
            missing_p1 = q1
            missing_p2 = 1 - q2
            frac1 = missing_p1 / (missing_p2 + missing_p1)
            frac2 = missing_p2 / (missing_p2 + missing_p1)
            missing_extent = extrap_total_extent - inner_extent

            pad1 = missing_extent * frac1
            pad2 = missing_extent * frac2

            low_ = low_ - pad1
            high_ = high_ + pad2

            low = low_ if low is None else min(low_, low)
            high = high_ if high is None else max(high_, high)
        return (low, high)

    # Hack values that we don't apply smoothing to
    HACK_NO_SMOOTH = ['lr', 'momentum']

    def tag_grouper(k):
        # parts = ['train_epoch', 'vali_epoch', 'test_epoch']
        # parts = [p.replace('epoch', 'mode') for p in parts]
        parts = [p + mode for p in ['train_', 'vali_', 'test_']]
        for p in parts:
            if p in k:
                return p.split('_')[0]
        return 'unknown'

    GROUP_LOSSES = True
    GROUP_AND_INDIVIDUAL = False
    INDIVIDUAL_PLOTS = True
    GROUP_SPECIAL = True

    if GROUP_LOSSES:
        # Group all losses in one plot for comparison
        loss_keys = [k for k in keys if 'loss' in k]
        tagged_losses = ub.group_items(loss_keys, tag_grouper)
        tagged_losses.pop('unknown', None)
        kw = {}
        kw['ymin'] = 0.0
        # print('tagged_losses = {!r}'.format(tagged_losses))
        for tag, losses in tagged_losses.items():

            min_abs_y = .01
            min_y = 0
            xydata = ub.odict()
            for key in sorted(losses):
                ydata = tb_data[key]['ydata']

                if HACK_NO_SMOOTH not in key.split('_'):
                    ydata = smooth_curve(ydata, smoothing)

                try:
                    min_y = min(min_y, ydata.min())
                    pos_ys = ydata[ydata > 0]
                    min_abs_y = min(min_abs_y, pos_ys.min())
                except Exception:
                    pass

                xydata[key] = (tb_data[key]['xdata'], ydata)

            kw['ymin'] = min_y

            if ignore_outliers:
                low, kw['ymax'] = inlier_ylim([t[1] for t in xydata.values()])

            yscales = ['symlog', 'linear']
            for yscale in yscales:
                fig.clf()
                ax = fig.gca()
                title = nice + '\n' + tag + '_' + mode + ' losses'
                kwplot.multi_plot(xydata=xydata, ylabel='loss', xlabel=mode,
                                  yscale=yscale, title=title, fnum=1, ax=ax,
                                  **kw)
                if yscale == 'symlog':
                    ax.set_yscale('symlog', linthreshy=min_abs_y)
                fname = '_'.join([tag, mode, 'multiloss', yscale]) + '.png'
                fpath = join(out_dpath, fname)
                ax.figure.savefig(fpath)

        # don't dump losses individually if we dump them in a group
        if not GROUP_AND_INDIVIDUAL:
            keys.difference_update(set(loss_keys))
            # print('keys = {!r}'.format(keys))

    if GROUP_SPECIAL:
        tag_groups = ub.group_items(keys, tag_grouper)
        tag_groups.pop('unknown', None)
        # Group items matching these strings
        kw = {}
        for tag, tag_keys in tag_groups.items():
            for groupname in special_groupers:
                group_keys = [k for k in tag_keys if groupname in k.split('_')]
                if len(group_keys) > 1:
                    # Gather data for this group
                    xydata = ub.odict()
                    for key in sorted(group_keys):
                        ydata = tb_data[key]['ydata']
                        if HACK_NO_SMOOTH not in key.split('_'):
                            ydata = smooth_curve(ydata, smoothing)
                        xydata[key] = (tb_data[key]['xdata'], ydata)

                    if ignore_outliers:
                        low, kw['ymax'] = inlier_ylim([t[1] for t in xydata.values()])

                    yscales = ['linear']
                    for yscale in yscales:
                        fig.clf()
                        ax = fig.gca()
                        title = nice + '\n' + tag + '_' + mode + ' ' + groupname
                        kwplot.multi_plot(xydata=xydata, ylabel=groupname, xlabel=mode,
                                          yscale=yscale, title=title, fnum=1, ax=ax,
                                          **kw)
                        if yscale == 'symlog':
                            ax.set_yscale('symlog', linthreshy=min_abs_y)
                        fname = '_'.join([tag, mode, 'group-' + groupname, yscale]) + '.png'
                        fpath = join(out_dpath, fname)
                        ax.figure.savefig(fpath)

                    if not GROUP_AND_INDIVIDUAL:
                        keys.difference_update(set(group_keys))

    if INDIVIDUAL_PLOTS:
        # print('keys = {!r}'.format(keys))
        for key in keys:
            d = tb_data[key]

            ydata = d['ydata']
            ydata = smooth_curve(ydata, smoothing)

            kw = {}
            if any(m.lower() in key.lower() for m in y01_measures):
                kw['ymin'] = 0.0
                kw['ymax'] = 1.0
            elif any(m.lower() in key.lower() for m in y0_measures):
                kw['ymin'] = min(0.0, ydata.min())
                if ignore_outliers:
                    low, kw['ymax'] = inlier_ylim([ydata])

            # NOTE: this is actually pretty slow
            fig.clf()
            ax = fig.gca()
            title = nice + '\n' + key
            kwplot.multi_plot(d['xdata'], ydata, ylabel=key, xlabel=mode,
                              title=title, fnum=1, ax=ax, **kw)

            # png is slightly smaller than jpg for this kind of plot
            fpath = join(out_dpath, key + '.png')
            # print('save fpath = {!r}'.format(fpath))
            ax.figure.savefig(fpath)



"""
Decode YOLO network outputs into netharn-style Detections

Adapted from code by EAVISE
"""
import torch
from netharn import util
import numpy as np  # NOQA
import ubelt as ub  # NOQA


class GetBoundingBoxes(object):
    """ Convert output from darknet networks to bounding box tensor.

    Args:
        network (lightnet.network.Darknet): Network the converter will be used with
        conf_thresh (Number [0-1]): Confidence threshold to filter detections
        nms_thresh(Number [0-1]): Overlapping threshold to filter detections with non-maxima suppresion

    Returns:
        List[netharn.util.Detections]:
            detection object for each image in the batch.

    Note:
        The boxes in the detections use relative values for its coordinates.

    Examples:
        >>> import torch
        >>> torch.random.manual_seed(0)
        >>> anchors = np.array([(1.3221, 1.73145), (3.19275, 4.00944), (5.05587, 8.09892), (9.47112, 4.84053), (11.2364, 10.0071)])
        >>> self = GetBoundingBoxes(anchors=anchors, num_classes=20, conf_thresh=.14, nms_thresh=0.5)
        >>> output = torch.randn(8, 5, 5 + 20, 9, 9)
        >>> batch_dets = self(output)
        >>> assert len(batch_dets) == 8
    """

    def __init__(self, num_classes, anchors, conf_thresh=0.001, nms_thresh=0.4):
        self.num_classes = num_classes
        self.num_anchors = len(anchors)
        self.anchor_step = len(anchors[0])
        self.anchors = torch.Tensor(anchors)
        self.conf_thresh = conf_thresh
        self.nms_thresh = nms_thresh

    @util.profile
    def __call__(self, output):
        """ Compute bounding boxes after thresholding and nms """
        batch_dets = self._decode(output.data)
        return batch_dets

    @util.profile
    def _decode(self, output):
        """
        Returns array of detections for every image in batch

        CommandLine:
            python ~/code/netharn/netharn/box_models/yolo2/light_postproc.py GetBoundingBoxes._decode

        Examples:
            >>> import torch
            >>> torch.random.manual_seed(0)
            >>> anchors = np.array([(1.3221, 1.73145), (3.19275, 4.00944), (5.05587, 8.09892), (9.47112, 4.84053), (11.2364, 10.0071)])
            >>> self = GetBoundingBoxes(anchors=anchors, num_classes=20, conf_thresh=.14, nms_thresh=0.5)
            >>> output = torch.randn(16, 5, 5 + 20, 9, 9)
            >>> from netharn import XPU
            >>> output = XPU.coerce('auto').move(output)
            >>> batch_dets = self._decode(output.data)
            >>> assert len(batch_dets) == 16

        Ignore:
            >>> from netharn.models.yolo2.yolo2 import *  # NOQA
            >>> info = dev_demodata()
            >>> outputs = info['outputs']
            >>> cxywh_energy = output['cxywh_energy']
            >>> raw = info['raw']
            >>> raw_ = raw.clone()

            >>> self = GetBoundingBoxes(anchors=info['model'].anchors, num_classes=20, conf_thresh=.14, nms_thresh=0.5)
            >>> dets = self._decode(raw)[0]
            >>> dets.scores

            >>> self, output = ub.take(info, ['coder', 'outputs'])
            >>> batch_dets = self.decode_batch(output)
            >>> dets = batch_dets[0]
            >>> dets.scores

        """
        # dont modify inplace
        raw_ = output.clone()

        # Variables
        bsize = raw_.shape[0]
        h, w = raw_.shape[-2:]

        device = raw_.device

        if self.anchors.device != device:
            self.anchors = self.anchors.to(device)

        # Compute xc,yc, w,h, box_score on Tensor
        lin_x = torch.linspace(0, w - 1, w, device=device).repeat(h, 1).view(h * w)
        lin_y = torch.linspace(0, h - 1, h, device=device).repeat(w, 1).t().contiguous().view(h * w)

        anchor_w = self.anchors[:, 0].contiguous().view(1, self.num_anchors, 1)
        anchor_h = self.anchors[:, 1].contiguous().view(1, self.num_anchors, 1)

        # -1 == 5+num_classes (we can drop feature maps if 1 class)
        output_ = raw_.view(bsize, self.num_anchors, -1, h * w)

        output_[:, :, 0, :].sigmoid_().add_(lin_x).div_(w)          # X center
        output_[:, :, 1, :].sigmoid_().add_(lin_y).div_(h)          # Y center
        output_[:, :, 2, :].exp_().mul_(anchor_w).div_(w)           # Width
        output_[:, :, 3, :].exp_().mul_(anchor_h).div_(h)           # Height
        output_[:, :, 4, :].sigmoid_()                              # Box score

        # output_[:, :, 0:4].sum()
        # torch.all(cxywh.view(-1) == output_[:, :, 0:4].contiguous().view(-1))

        # Compute class_score
        if self.num_classes > 1:
            cls_scores = torch.nn.functional.softmax(output_[:, :, 5:, :], 2)

            cls_max, cls_max_idx = torch.max(cls_scores, 2)
            cls_max.mul_(output_[:, :, 4, :])
        else:
            cls_max = output_[:, :, 4, :]
            cls_max_idx = torch.zeros_like(cls_max)

        # Save detection if conf*class_conf is higher than threshold

        # Newst lightnet code, which is based on my mode1 code
        score_thresh = cls_max > self.conf_thresh
        score_thresh_flat = score_thresh.view(-1)

        if score_thresh.sum() == 0:
            batch_dets = []
            for i in range(bsize):
                batch_dets.append(util.Detections(
                    boxes=util.Boxes(torch.empty((0, 4), dtype=torch.float32, device=device), 'cxywh'),
                    scores=torch.empty(0, dtype=torch.float32, device=device),
                    class_idxs=torch.empty(0, dtype=torch.int64, device=device),
                ))
        else:
            # Mask select boxes > conf_thresh
            coords = output_.transpose(2, 3)[..., 0:4]
            coords = coords[score_thresh[..., None].expand_as(coords)].view(-1, 4)

            scores = cls_max[score_thresh]

            class_idxs = cls_max_idx[score_thresh]

            stacked_dets = util.Detections(
                boxes=util.Boxes(coords, 'cxywh'),
                scores=scores,
                class_idxs=class_idxs,
            )

            # Get indexes of splits between images of batch
            max_det_per_batch = len(self.anchors) * h * w
            slices = [slice(max_det_per_batch * i, max_det_per_batch * (i + 1)) for i in range(bsize)]
            det_per_batch = torch.IntTensor([score_thresh_flat[s].int().sum() for s in slices])
            split_idx = torch.cumsum(det_per_batch, dim=0)

            batch_dets = []
            start = 0
            for end in split_idx:
                dets = stacked_dets[start: end]
                dets = dets.non_max_supress(thresh=self.nms_thresh)
                batch_dets.append(dets)
                start = end
        return batch_dets


if __name__ == '__main__':
    """
    CommandLine:
        python -m netharn.models.yolo2.light_postproc all
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
References:
    https://arxiv.org/pdf/1809.02983.pdf - Dual Attention Network for Scene Segmentation

    https://raw.githubusercontent.com/heykeetae/Self-Attention-GAN/master/sagan_models.py
"""
import torch
from torch import nn


class SelfAttention(nn.Module):
    """
    Self Attention Layer

    References:
    """

    def __init__(self, in_channels):
        super(SelfAttention, self).__init__()
        self.chanel_in = in_channels

        self.query_conv = nn.Conv2d(in_channels=in_channels,
                                    out_channels=in_channels // 8,
                                    kernel_size=1)

        self.key_conv = nn.Conv2d(in_channels=in_channels,
                                  out_channels=in_channels // 8, kernel_size=1)

        self.value_conv = nn.Conv2d(in_channels=in_channels,
                                    out_channels=in_channels, kernel_size=1)

        self.gamma = nn.Parameter(torch.zeros(1))

        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x):
        """
        Args :
            x (Tensor): input feature maps (B x C x W x H)

        Returns :
            out : self attention value + input feature
            attention: B x N x N (N is Width*Height)
        """
        B, C, W, H = x.shape
        N = W * H

        proj_query = self.query_conv(x).view(B, -1, N).permute(0, 2, 1)  # B x C x(N)

        proj_key = self.key_conv(x).view(B, -1, N)  # B x C x (*W*H)

        energy = torch.bmm(proj_query, proj_key)  # transpose check

        attention = self.softmax(energy)  # B x (N) x (N)

        proj_value = self.value_conv(x).view(B, -1, N)  # B x C x N

        out = torch.bmm(proj_value, attention.permute(0, 2, 1))
        out = out.view(B, C, W, H)

        out = self.gamma * out + x
        return out, attention


class ChannelAttention(nn.Module):
    """
    Channel attention module

    The channel attention module selectively emphasizes interdependent channel
    maps by integrating associated features among all channel map.

    Uses the uncentered scatter matrix (i.e. M @ M.T) to compute a unnormalized
    correlation-like matrix between channels.

    I think M @ M.T is an "uncentered scatter matrix"

    https://stats.stackexchange.com/questions/164997/relationship-between-gram-and-covariance-matrices

    not sure if this is the right term

    References:
        https://arxiv.org/pdf/1809.02983.pdf - Dual Attention Network for Scene Segmentation
        https://github.com/junfu1115/DANet/blob/master/encoding/nn/attention.py

    Notes:
         Different from the position attention module, we directly calculate
         the channel attention map from the original features.

         Noted that we do not employ convolution layers to embed features
         before computing relationshoips of two channels, since it can maintain
         relationship between different channel maps. In addition, different
         from recent works [Zhang CVPR 2018 Context encoding for semantic
         segmentation] which explores channel relationships by a global pooling
         or encoding layer, we exploit spatial information at all corresponding
         positions to model channel correlations

    Ignore:

        >>> # Simple example to demonstrate why a multiplicative parameter
        >>> # at zero might or might not deviate to decrease the loss
        >>> x = torch.randn(10)
        >>> x[0] = -1000
        >>> p = nn.Parameter(torch.zeros(1) + 1e-1)
        >>> optim = torch.optim.SGD([p], lr=1e-1)
        >>> for i in range(10):
        >>>     loss = (x * (p ** 2)).sum()
        >>>     loss.backward()
        >>>     print('loss = {!r}'.format(loss))
        >>>     print('p.data = {!r}'.format(p.data))
        >>>     print('p.grad = {!r}'.format(p.grad))
        >>>     optim.step()
        >>>     optim.zero_grad()

        >>> # at zero might or might not deviate to decrease the loss
        >>> x = torch.randn(2)
        >>> x[0] = -1000
        >>> p = nn.Parameter(torch.zeros(1))
        >>> optim = torch.optim.SGD([p], lr=1e-1)
        >>> for i in range(10):
        >>>     loss = (x * p.clamp(0, None)).sum()
        >>>     loss.backward()
        >>>     print('loss = {!r}'.format(loss))
        >>>     print('p.data = {!r}'.format(p.data))
        >>>     print('p.grad = {!r}'.format(p.grad))
        >>>     optim.step()
        >>>     optim.zero_grad()

    Ignore:
        >>> B, C, H, W = 1, 3, 5, 7
        >>> inputs = torch.rand(B, C, H, W)
        >>> inputs = torch.arange(B * C * H * W).view(B, C, H, W).float()
        >>> self = ChannelAttention(C)
        >>> optim = torch.optim.SGD(self.parameters(), lr=1e-8)
        >>> for i in range(10):
        >>>     out = self(inputs)
        >>>     loss = (out.sum() ** 2)
        >>>     print('self.gamma = {!r}'.format(self.gamma))
        >>>     print('loss = {!r}'.format(loss))
        >>>     loss.backward()
        >>>     optim.step()
        >>>     optim.zero_grad()
    """
    def __init__(self, in_channels, attend_elsewhere=True):
        super(ChannelAttention, self).__init__()
        self.in_channels = in_channels

        # hack to rectify the definiton in the paper with the implementaiton
        self.attend_elsewhere = attend_elsewhere

        # scale parameter (beta from paper)
        self.gamma = nn.Parameter(torch.zeros(1))

    def forward(self, inputs):
        """
        Args:
            inputs (Tensor): input feature maps (B, C, H, W)

        Returns:
            out (Tensor): attention value + input feature
            attention: (B, C, C)

        Example:
            >>> B, C, H, W = 1, 3, 5, 7
            >>> inputs = torch.rand(B, C, H, W)
            >>> self = ChannelAttention(C)
        """
        B, C, H, W = inputs.shape

        # Flatten spatial dims
        proj_query = inputs.view(B, C, -1)  # A
        proj_key = inputs.view(B, C, -1).permute(0, 2, 1)  # A.T
        proj_value = inputs.view(B, C, -1)  # A

        energy = torch.bmm(proj_query, proj_key)  # A @ A.T

        if self.attend_elsewhere:
            # Why the subtraction here?
            diag = torch.max(energy, dim=1, keepdim=True)[0].expand_as(energy)
            energy_new = diag - energy
            attention = energy_new.softmax(dim=1)
        else:
            attention = energy.softmax(dim=1)

        out = torch.bmm(attention, proj_value)
        out = out.view(B, C, H, W)

        residual = self.gamma * out
        out = residual + inputs
        return out



import torch.nn.functional as F
import torch
from torch import nn
import ubelt as ub
from netharn.layers import common
from netharn.analytic import output_shape_for
from netharn.analytic import receptive_field_for


class L2Norm(common.Module):
    """
    L2Norm layer across all channels

    Notes:
        "The L2 normalization technique introduced in [12] to scale the feature
        norm at each location in the feature map to `scale` and learn the scale
        during back propagation."

        * In my experience (author of netharn), using this has tended to hurt
        more than help.

    References:
        [12] Liu, Rabinovich, Berg - ParseNet: Looking wider to see better (ILCR) (2016)

    Example:
        >>> import numpy as np
        >>> import ubelt as ub
        >>> in_features = 7
        >>> self = L2Norm(in_features, scale=20)
        >>> x = torch.rand(1, in_features, 2, 2)
        >>> y = self(x)
        >>> norm = np.linalg.norm(y.data.cpu().numpy(), axis=1)
        >>> print(ub.repr2(norm, precision=2))
        np.array([[[20., 20.],
                   [20., 20.]]], dtype=np.float32)

    Example:
        >>> from netharn.analytic.output_shape_for import OutputShapeFor
        >>> self = L2Norm(in_features=7, scale=20)
        >>> OutputShapeFor(self)._check_consistency((1, 7, 2, 2))
        (1, 7, 2, 2)
    """

    def __init__(self, in_features, scale):
        super(L2Norm, self).__init__()
        self.weight = nn.Parameter(torch.Tensor(in_features))
        self._initial_scale = scale
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.constant_(self.weight, self._initial_scale)

    def forward(self, x):
        x = F.normalize(x, dim=1)
        scale = self.weight[None, :, None, None]
        return scale * x

    def output_shape_for(self, input_shape):
        return output_shape_for.OutputShape.coerce(input_shape)

    def receptive_field_for(self, input_field=None):
        return input_field


class InputNorm(common.Module):
    """
    Normalizes the input by shifting and dividing by a scale factor.

    This allows for the network to take care of 0-mean 1-std normalization.
    The developer explicitly specifies what these shift and scale values are.
    By specifying this as a layer (instead of a data preprocessing step), the
    netharn exporter will remember and associated this information with any
    deployed model. This means that a user does not need to remember what these
    shit/scale arguments were before passing inputs to a network.

    If the mean and std arguments are unspecified, this layer becomes a noop.

    References:
        https://towardsdatascience.com/why-data-should-be-normalized-before-training-a-neural-network-c626b7f66c7d

    Example:
        >>> self = InputNorm(mean=50.0, std=29.0)
        >>> inputs = torch.rand(2, 3, 5, 7) * 100
        >>> outputs = self(inputs)
        >>> # If mean and std are unspecified, this becomes a noop.
        >>> assert torch.all(InputNorm()(inputs) == inputs)
        >>> # Specifying either the mean or the std is ok.
        >>> partial1 = InputNorm(mean=50)(inputs)
        >>> partial2 = InputNorm(std=29)(inputs)
    """

    def __init__(self, mean=None, std=None):
        super(InputNorm, self).__init__()
        if mean is not None:
            mean = mean if ub.iterable(mean) else [mean]
            mean = torch.FloatTensor(mean)
        if std is not None:
            std = std if ub.iterable(std) else [std]
            std = torch.FloatTensor(std)
        self.register_buffer('mean', mean)
        self.register_buffer('std', std)

    def forward(self, inputs):
        outputs = inputs
        if self.mean is not None:
            outputs = outputs - self.mean
        if self.std is not None:
            outputs = outputs / self.std
        return outputs

    def output_shape_for(self, input_shape):
        return output_shape_for.OutputShape.coerce(input_shape)

    def receptive_field_for(self, input_field=None):
        return receptive_field_for.ReceptiveField.coerce(input_field)



# coding=utf8
import math
import numpy as np
import torch
from netharn.layers import common
import torch.nn.functional as F


class Conv1d_pad(torch.nn.Conv1d, common.ModuleMixin):
    """
    Extends convolutions to include the type of mode used in padding
    """
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True, mode='reflect'):
        super(Conv1d_pad, self).__init__(in_channels, out_channels,
                                         kernel_size, stride=stride, padding=0,
                                         dilation=dilation, groups=groups,
                                         bias=bias)
        self._custom_padding = padding
        self._pad_mode = mode
        self.ndim = 1

    def forward(self, input):
        # Custom pad, then convolve
        x = F.pad(input, [self._custom_padding] * (2 * self.ndim), mode=self._pad_mode)
        return super(Conv1d_pad, self).forward(x)


class Conv2d_pad(torch.nn.Conv2d, common.ModuleMixin):
    """
    Extends convolutions to include the type of mode used in padding
    """
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True, mode='reflect'):
        super(Conv2d_pad, self).__init__(in_channels, out_channels,
                                         kernel_size, stride=stride, padding=0,
                                         dilation=dilation, groups=groups,
                                         bias=bias)
        self._custom_padding = padding
        self._pad_mode = mode
        self.ndim = 2

    def forward(self, input):
        # Custom pad, then convolve
        x = F.pad(input, [self._custom_padding] * (2 * self.ndim), mode=self._pad_mode)
        return super(Conv2d_pad, self).forward(x)


class GaussianBlurNd(common.Module):
    """
    Convolves a signal with a Gaussian kernel.

    Args:
        dim (int): number of input space-time dims (ignore batch / channel)

        num_features (int): number of incoming channels

        sigma (float): standard deviation of Gaussian kernel.

        truncate (float): ignored if kernel_size is not None, otherwise
            computes kernel size that cuts off the filter at this many standard
            deviations. (See scipy.ndimage.gaussian_filter1d)

        kernel_size (int): discrete window size (if not given, it is computed
            based on sigma and truncate)

        separable (bool): if True uses n seperable 1d convolutions instead of
            a single Nd convolution. This usually results in a speedup.
            Defaults to True. Note: the False case is only implemented for
            `dim=2`.

    Example:
        >>> import netharn as nh
        >>> from torchvision.transforms.functional import to_tensor
        >>> input = to_tensor(nh.util.grab_test_image())[None, :]
        >>> self = GaussianBlurNd(2, num_features=input.shape[1], sigma=1.4)
        >>> assert self.number_of_parameters() == 0
        >>> output = self(input)
        >>> # xdoc: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(input[0], colorspace='rgb', pnum=(1, 2, 1))
        >>> kwplot.imshow(output[0], colorspace='rgb', pnum=(1, 2, 2))
        >>> kwplot.show_if_requested()

    References:
        https://discuss.pytorch.org/t/is-there-anyway-to-do-gaussian-filtering-for-an-image-2d-3d-in-pytorch/12351

    Benchmark:
        import netharn as nh
        import cv2
        import ubelt as ub
        from torchvision.transforms.functional import to_tensor

        image = nh.util.grab_test_image('carl')
        image = cv2.resize(image, (256, 256))
        im = to_tensor(image)
        input = torch.stack([im] * 16, dim=0)

        self1 = GaussianBlurNd(2, 3, sigma=math.sqrt(2), separable=True)
        self2 = GaussianBlurNd(2, 3, sigma=math.sqrt(2), separable=False)

        xpu = nh.XPU(0)
        self1, self2, input = map(xpu.move, [self1, self2, input])

        # Applying the 1d Gaussian multiple times has a clear time advantage
        ti = ub.Timerit(100, bestof=10, verbose=1)
        for timer in ti.reset('time-1d'):
            with timer:
                self1(input)
                torch.cuda.synchronize()
        # Timed best=2.162 ms, mean=2.167 ± 0.0 ms for time-1d

        for timer in ti.reset('time-2d'):
            with timer:
                self2(input)
                torch.cuda.synchronize()
        # Timed best=5.998 ms, mean=6.082 ± 0.1 ms for time-2d
    """

    def __init__(self, dim, num_features, sigma, kernel_size=None,
                 truncate=4.0, separable=False):
        super(GaussianBlurNd, self).__init__()
        self.separable = separable

        # Create a x, y coordinate grid of shape (kernel_size, kernel_size, 2)
        # truncate = 4
        if kernel_size is None:
            lw = int(truncate * sigma + 0.5)
            kernel_size = lw * 2 + 1
        else:
            assert kernel_size % 2 == 1
            lw = (kernel_size - 1) // 2

        if self.separable:
            # Calculate the 1d Gaussian kernel
            # Follow scipy.ndimage method closely
            import scipy
            import scipy.ndimage
            kernel1d = scipy.ndimage.filters._gaussian_kernel1d(sigma, order=0, radius=lw)[::-1]
            kernel1d = torch.from_numpy(np.ascontiguousarray(kernel1d))

            # Reshape to 1d depthwise convolutional weight
            kernel1d = kernel1d.view(1, kernel_size)
            kernel1d = kernel1d.repeat(num_features, 1, 1)

            self.padding = lw

            gauss1d = Conv1d_pad(in_channels=num_features, out_channels=num_features,
                                 kernel_size=kernel_size, groups=num_features,
                                 bias=False, padding=self.padding)

            gauss1d.weight.data[:] = kernel1d
            gauss1d.weight.requires_grad = False
            self.gauss1d = gauss1d
        else:
            kernel_size = lw * 2 + 1
            mean = (kernel_size - 1) / 2.0
            x_cord = torch.arange(kernel_size)
            x_grid = x_cord.repeat(kernel_size).view(kernel_size, kernel_size)
            y_grid = x_grid.t()
            xy_grid = torch.stack([x_grid, y_grid], dim=-1).float()

            coeff = (1.0 / (2.0 * math.pi * sigma ** 2.0))
            exponent = -torch.sum((xy_grid - mean) ** 2.0, dim=-1) / (2 * sigma ** 2.0)
            kernel2d = coeff * torch.exp(exponent)
            kernel2d /= torch.sum(kernel2d)

            # Reshape to 2d depthwise convolutional weight
            kernel2d = kernel2d.view(1, 1, kernel_size, kernel_size)
            kernel2d = kernel2d.repeat(num_features, 1, 1, 1)

            gauss2d = Conv2d_pad(in_channels=num_features, out_channels=num_features,
                                 kernel_size=kernel_size, groups=num_features,
                                 bias=False, padding=kernel_size // 2)

            gauss2d.weight.data[:] = kernel2d
            gauss2d.weight.requires_grad = False
            self.gauss2d = gauss2d

    def forward(self, input):
        if self.separable:
            # Apply the 1d gaussian filter in each dimension
            b, c = input.shape[0:2]
            dims = input.shape[2:]
            x = input
            major_dim = 2  # tensor dim corresponding to major spatial axis
            for axis, d in enumerate(dims):
                if axis > 0:
                    # Transpose to put the axis of interest in front
                    x = x.transpose(major_dim, major_dim + axis).contiguous()
                # Flatten so current axis is stacked
                shape = x.shape
                axis_view = x.view(b, c, -1)
                axis_view = self.gauss1d.forward(axis_view)
                x = axis_view.view(*shape)
                if axis > 0:
                    # Restore original axis order
                    x = x.transpose(major_dim, major_dim + axis)
            output = x.contiguous()
            return output
        else:
            return self.gauss2d.forward(input)

    def output_shape_for(self, input_shape):
        return input_shape

    def receptive_field_for(self, input_field=None):
        # Even though this does change the receptive feild a little bit
        # lets pretend that it doesnt
        return input_field



from os.path import relpath
from os.path import exists
from os.path import join
import ubelt as ub


def _devcheck_sample_full_image():
    """
    """
    import kwimage
    import numpy as np

    sampler = grab_camvid_sampler()

    cid_to_cidx = sampler.catgraph.id_to_idx
    classes = sampler.catgraph

    # Try loading an entire image
    img, annots = sampler.load_image_with_annots(1)

    file = img['imdata']
    imdata = file[:]

    aids = [ann['id'] for ann in annots]
    _annots = sampler.dset.annots(aids)

    sseg_list = []
    for s in _annots.lookup('segmentation'):
        m = kwimage.MultiPolygon.coerce(s)
        sseg_list.append(m)

    aids = _annots.aids
    cids = _annots.cids
    boxes = _annots.boxes
    segmentations = kwimage.PolygonList(sseg_list)
    class_idxs = np.array([cid_to_cidx[cid] for cid in cids])

    dets = kwimage.Detections(
        aids=aids,
        boxes=boxes,
        class_idxs=class_idxs,
        segmentations=segmentations,
        classes=classes,
        datakeys=['aids'],
    )

    if 1:
        print('dets = {!r}'.format(dets))
        print('dets.data = {!r}'.format(dets.data))
        print('dets.meta = {!r}'.format(dets.meta))

    if ub.argflag('--show'):
        import kwplot

        with ub.Timer('dets.draw_on'):
            canvas = imdata.copy()
            canvas = dets.draw_on(canvas)
            kwplot.imshow(canvas, pnum=(1, 2, 1), title='dets.draw_on')

        with ub.Timer('dets.draw'):
            kwplot.imshow(imdata, pnum=(1, 2, 2), docla=True, title='dets.draw')
            dets.draw()


def _devcheck_load_sub_image():
    import kwimage
    import numpy as np

    sampler = grab_camvid_sampler()

    cid_to_cidx = sampler.catgraph.id_to_idx
    classes = sampler.catgraph

    # Try loading a subregion of an image
    sample = sampler.load_positive(2)
    imdata = sample['im']
    annots = sample['annots']
    aids = annots['aids']
    cids = annots['cids']
    boxes = annots['rel_boxes']
    class_idxs = np.array([cid_to_cidx[cid] for cid in cids])
    segmentations = annots['rel_ssegs']

    raw_dets = kwimage.Detections(
        aids=aids,
        boxes=boxes,
        class_idxs=class_idxs,
        segmentations=segmentations,
        classes=classes,
        datakeys=['aids'],
    )

    # Clip boxes to the image boundary
    input_dims = imdata.shape[0:2]
    raw_dets.data['boxes'] = raw_dets.boxes.clip(0, 0, input_dims[1], input_dims[0])

    keep = []
    for i, s in enumerate(raw_dets.data['segmentations']):
        # TODO: clip polygons
        m = s.to_mask(input_dims)
        if m.area > 0:
            keep.append(i)
    dets = raw_dets.take(keep)

    heatmap = dets.rasterize(bg_size=(1, 1), input_dims=input_dims)

    if 1:
        print('dets = {!r}'.format(dets))
        print('dets.data = {!r}'.format(dets.data))
        print('dets.meta = {!r}'.format(dets.meta))

    if ub.argflag('--show'):
        import kwplot

        kwplot.autompl()
        heatmap.draw()

        draw_boxes = 1

        kwplot.figure(doclf=True)
        with ub.Timer('dets.draw_on'):
            canvas = imdata.copy()
            # TODO: add logic to color by class
            canvas = dets.draw_on(canvas, boxes=draw_boxes, color='random')
            kwplot.imshow(canvas, pnum=(1, 2, 1), title='dets.draw_on')

        with ub.Timer('dets.draw'):
            kwplot.imshow(imdata, pnum=(1, 2, 2), docla=True, title='dets.draw')
            dets.draw(boxes=draw_boxes, color='random')


def grab_camvid_train_test_val_splits(coco_dset, mode='segnet'):
    # Use the split from SegNet: https://github.com/alexgkendall/SegNet-Tutorial
    split_files = {
        'train': ub.grabdata('https://raw.githubusercontent.com/alexgkendall/SegNet-Tutorial/master/CamVid/train.txt'),
        'vali': ub.grabdata('https://raw.githubusercontent.com/alexgkendall/SegNet-Tutorial/master/CamVid/val.txt'),
        'test': ub.grabdata('https://raw.githubusercontent.com/alexgkendall/SegNet-Tutorial/master/CamVid/test.txt'),
    }
    gid_subsets = {}
    for tag, fpath in split_files.items():
        text = open(fpath, 'r').read()
        parts = text.replace('\n', ' ').split(' ')
        parts = [p for p in parts if p]
        from os.path import basename
        names = sorted(set(basename(p) for p in parts))
        gids = [coco_dset.index.file_name_to_img['701_StillsRaw_full/' + name]['id'] for name in names]
        gid_subsets[tag] = gids
    return gid_subsets


def grab_camvid_sampler():
    """
    Grab a ndsampler.CocoSampler object for the CamVid dataset.

    Returns:
        ndsampler.CocoSampler: sampler

    Example:
        >>> # xdoctest: +REQUIRES(--download)
        >>> sampler = grab_camvid_sampler()
        >>> print('sampler = {!r}'.format(sampler))
        >>> # sampler.load_sample()
        >>> for gid in ub.ProgIter(sampler.image_ids, desc='load image'):
        >>>     img = sampler.load_image(gid)
    """
    import ndsampler
    dset = grab_coco_camvid()
    workdir = ub.ensure_app_cache_dir('camvid')
    sampler = ndsampler.CocoSampler(dset, workdir=workdir)
    return sampler


def grab_coco_camvid():
    """
    Example:
        >>> # xdoctest: +REQUIRES(--download)
        >>> dset = grab_coco_camvid()
        >>> print('dset = {!r}'.format(dset))
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> plt = kwplot.autoplt()
        >>> plt.clf()
        >>> dset.show_image(gid=1)

    Ignore:
        import xdev
        gid_list = list(dset.imgs)
        for gid in xdev.InteractiveIter(gid_list):
            dset.show_image(gid)
            xdev.InteractiveIter.draw()
    """
    import ndsampler
    cache_dpath = ub.ensure_app_cache_dir('netharn', 'camvid')
    coco_fpath = join(cache_dpath, 'camvid.mscoco.json')

    # Need to manually bump this if you make a change to loading
    SCRIPT_VERSION = 'v4'

    # Ubelt's stamp-based caches are super cheap and let you take control of
    # the data format.
    stamp = ub.CacheStamp('camvid_coco', cfgstr=SCRIPT_VERSION,
                          dpath=cache_dpath, product=coco_fpath, hasher='sha1',
                          verbose=3)
    if stamp.expired():
        camvid_raw_info = grab_raw_camvid()
        dset = convert_camvid_raw_to_coco(camvid_raw_info)
        with ub.Timer('dumping MS-COCO dset to: {}'.format(coco_fpath)):
            dset.dump(coco_fpath)
        # Mark this process as completed by saving a small file containing the
        # hash of the "product" you are stamping.
        stamp.renew()

    # We can also cache the index build step independently. This uses
    # ubelt.Cacher, which is pickle based, and writes the actual object to
    # disk. Each type of caching has its own uses and tradeoffs.
    cacher = ub.Cacher('prebuilt-coco', cfgstr=SCRIPT_VERSION,
                       dpath=cache_dpath, verbose=3)
    dset = cacher.tryload()
    if dset is None:
        print('Reading coco_fpath = {!r}'.format(coco_fpath))
        dset = ndsampler.CocoDataset(coco_fpath, tag='camvid')
        # Directly save the file to disk.
        dset._build_index()
        dset._build_hashid()
        cacher.save(dset)

    camvid_dset = dset
    print('Loaded camvid_dset = {!r}'.format(camvid_dset))
    return camvid_dset


def grab_raw_camvid():
    """
    Grab the raw camvid data.
    """
    import zipfile
    dpath = ub.get_app_cache_dir('netharn', 'camvid')

    # url = 'http://mi.eng.cam.ac.uk/research/projects/VideoRec/CamVid/data/LabeledApproved_full.zip'
    # url = 'https://github.com/mostafaizz/camvid/archive/master.zip'
    url = 'https://data.kitware.com/api/v1/item/5cc0adce8d777f072b643503/download'
    zip_fpath = ub.grabdata(url, fname='camvid-master.zip', dpath=dpath)

    dset_root = join(dpath, 'camvid-master')
    image_dpath = join(dset_root, '701_StillsRaw_full')
    mask_dpath = join(dset_root, 'LabeledApproved_full')
    label_path = join(dset_root, 'label_colors.txt')

    if not exists(image_dpath):
        zip_ref = zipfile.ZipFile(zip_fpath, 'r')
        zip_ref.extractall(dpath)
        zip_ref.close()

    import glob
    img_paths = sorted([relpath(fpath, dset_root)
                        for fpath in glob.glob(join(image_dpath, '*.png'))])
    mask_paths = sorted([relpath(fpath, dset_root)
                         for fpath in glob.glob(join(mask_dpath, '*.png'))])

    camvid_raw_info = {
        'img_paths': img_paths,
        'mask_paths': mask_paths,
        'dset_root': dset_root,
        'label_path': label_path,
    }
    return camvid_raw_info


def rgb_to_cid(r, g, b):
    cid = (r << 16) + (g << 8) + (b << 0)
    return cid


def cid_to_rgb(cid):
    mask_b = (int(2 ** 8) - 1) << 0
    mask_g = (int(2 ** 8) - 1) << 8
    mask_r = (int(2 ** 8) - 1) << 16

    r = (cid & mask_b) >> 0
    g = (cid & mask_g) >> 8
    b = (cid & mask_r) >> 16
    rgb = (r, g, b)
    return rgb


def convert_camvid_raw_to_coco(camvid_raw_info):
    """
    Converts the raw camvid format to an MSCOCO based format, ( which lets use
    use ndsampler's COCO backend).

    Example:
        >>> # xdoctest: +REQUIRES(--download)
        >>> camvid_raw_info = grab_raw_camvid()
        >>> # test with a reduced set of data
        >>> del camvid_raw_info['img_paths'][2:]
        >>> del camvid_raw_info['mask_paths'][2:]
        >>> dset = convert_camvid_raw_to_coco(camvid_raw_info)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> plt = kwplot.autoplt()
        >>> kwplot.figure(fnum=1, pnum=(1, 2, 1))
        >>> dset.show_image(gid=1)
        >>> kwplot.figure(fnum=1, pnum=(1, 2, 2))
        >>> dset.show_image(gid=2)
    """
    import re
    import kwimage
    import ndsampler
    print('Converting CamVid to MS-COCO format')

    dset_root, img_paths, label_path, mask_paths = ub.take(
        camvid_raw_info, 'dset_root, img_paths, label_path, mask_paths'.split(', '))

    img_infos = {
        'img_fname': img_paths,
        'mask_fname': mask_paths,
    }
    keys = list(img_infos.keys())
    next_vals = list(zip(*img_infos.values()))
    image_items = [{k: v for k, v in zip(keys, vals)} for vals in next_vals]

    dataset = {
        'img_root': dset_root,
        'images': [],
        'categories': [],
        'annotations': [],
    }

    lines = ub.readfrom(label_path).split('\n')
    lines = [line for line in lines if line]
    for line in lines:
        color_text, name = re.split('\t+', line)
        r, g, b = map(int, color_text.split(' '))
        color = (r, g, b)

        # Parse the special camvid format
        cid = (r << 16) + (g << 8) + (b << 0)
        cat = {
            'id': cid,
            'name': name,
            'color': color,
        }
        dataset['categories'].append(cat)

    for gid, img_item in enumerate(image_items, start=1):
        img = {
            'id': gid,
            'file_name': img_item['img_fname'],
            # nonstandard image field
            'segmentation': img_item['mask_fname'],
        }
        dataset['images'].append(img)

    dset = ndsampler.CocoDataset(dataset)
    dset.rename_categories({'Void': 'background'})

    assert dset.name_to_cat['background']['id'] == 0
    dset.name_to_cat['background'].setdefault('alias', []).append('Void')

    if False:
        _define_camvid_class_hierarcy(dset)

    if 1:
        # TODO: Binarize CCs (and efficiently encode if possible)
        import numpy as np

        bad_info = []
        once = False

        # Add images
        dset.remove_all_annotations()
        for gid, img in ub.ProgIter(dset.imgs.items(), desc='parse label masks'):
            mask_fpath = join(dset_root, img['segmentation'])

            rgb_mask = kwimage.imread(mask_fpath, space='rgb')
            r, g, b  = rgb_mask.T.astype(np.int64)
            cid_mask = np.ascontiguousarray(rgb_to_cid(r, g, b).T)

            cids = set(np.unique(cid_mask)) - {0}

            for cid in cids:
                if cid not in dset.cats:
                    if gid == 618:
                        # Handle a known issue with image 618
                        c_mask = (cid == cid_mask).astype(np.uint8)
                        total_bad = c_mask.sum()
                        if total_bad < 32:
                            if not once:
                                print('gid 618 has a few known bad pixels, ignoring them')
                                once = True
                            continue
                        else:
                            raise Exception('more bad pixels than expected')
                    else:
                        raise Exception('UNKNOWN cid = {!r} in gid={!r}'.format(cid, gid))

                    # bad_rgb = cid_to_rgb(cid)
                    # print('bad_rgb = {!r}'.format(bad_rgb))
                    # print('WARNING UNKNOWN cid = {!r} in gid={!r}'.format(cid, gid))
                    # bad_info.append({
                    #     'gid': gid,
                    #     'cid': cid,
                    # })
                else:
                    ann = {
                        'category_id': cid,
                        'image_id': gid
                        # 'segmentation': mask.to_coco()
                    }
                    assert cid in dset.cats
                    c_mask = (cid == cid_mask).astype(np.uint8)
                    mask = kwimage.Mask(c_mask, 'c_mask')

                    box = kwimage.Boxes([mask.get_xywh()], 'xywh')
                    # box = mask.to_boxes()

                    ann['bbox'] = ub.peek(box.to_coco())
                    ann['segmentation'] = mask.to_coco()
                    dset.add_annotation(**ann)

        if 0:
            bad_cids = [i['cid'] for i in bad_info]
            print(sorted([c['color'] for c in dataset['categories']]))
            print(sorted(set([cid_to_rgb(i['cid']) for i in bad_info])))

            gid = 618
            img = dset.imgs[gid]
            mask_fpath = join(dset_root, img['segmentation'])
            rgb_mask = kwimage.imread(mask_fpath, space='rgb')
            r, g, b  = rgb_mask.T.astype(np.int64)
            cid_mask = np.ascontiguousarray(rgb_to_cid(r, g, b).T)
            cid_hist = ub.dict_hist(cid_mask.ravel())

            bad_cid_hist = {}
            for cid in bad_cids:
                bad_cid_hist[cid] = cid_hist.pop(cid)

            import kwplot
            kwplot.autompl()
            kwplot.imshow(rgb_mask)

    if 0:
        import kwplot
        plt = kwplot.autoplt()
        plt.clf()
        dset.show_image(1)

        import xdev
        gid_list = list(dset.imgs)
        for gid in xdev.InteractiveIter(gid_list):
            dset.show_image(gid)
            xdev.InteractiveIter.draw()

    dset._build_index()
    dset._build_hashid()
    return dset


def _define_camvid_class_hierarcy(dset):
    # add extra supercategories
    # NOTE: life-conscious, and life-inanimate are disjoint in this
    # forumlation because we are restricted to a tree structure.  If
    # this changse, then we can try rencoding with multiple parents.
    extra_structure = {
        # Break down the image into things that are part of the system, and
        # things that aren't
        'background': 'root',
        'system': 'root',

        # The system is made up of environmental components and actor
        # components.
        'environment': 'system',
        'actor': 'system',

        # Break actors (things with complex movement) into subtypes
        'life-conscious': 'actor',
        'vehicle-land': 'actor',
        'actor-other': 'actor',

        # Break the environment (things with simple movement) info subtypes
        'life-inanimate': 'environment',
        'civil-structure': 'environment',
        'civil-notice': 'environment',
        'transport-way': 'environment',

        # Subclassify transport mediums
        'drive-way': 'transport-way',
        'walk-way': 'transport-way',
    }

    for child, parent in extra_structure.items():
        if child in dset.name_to_cat:
            dset.name_to_cat[child]['supercategory'] = parent
        else:
            dset.add_category(name=child, supercategory=parent)

    dset.name_to_cat['background']['supercategory'] = 'root'

    dset.name_to_cat['Sky']['supercategory'] = 'environment'

    dset.name_to_cat['Animal']['supercategory'] = 'life-conscious'
    dset.name_to_cat['Bicyclist']['supercategory'] = 'life-conscious'
    dset.name_to_cat['Pedestrian']['supercategory'] = 'life-conscious'
    dset.name_to_cat['Child']['supercategory'] = 'life-conscious'

    dset.name_to_cat['OtherMoving']['supercategory'] = 'actor-other'
    dset.name_to_cat['CartLuggagePram']['supercategory'] = 'actor-other'

    dset.name_to_cat['Car']['supercategory'] = 'vehicle-land'
    dset.name_to_cat['Train']['supercategory'] = 'vehicle-land'
    dset.name_to_cat['Truck_Bus']['supercategory'] = 'vehicle-land'
    dset.name_to_cat['SUVPickupTruck']['supercategory'] = 'vehicle-land'
    dset.name_to_cat['MotorcycleScooter']['supercategory'] = 'vehicle-land'

    dset.name_to_cat['VegetationMisc']['supercategory'] = 'life-inanimate'
    dset.name_to_cat['Tree']['supercategory'] = 'life-inanimate'

    dset.name_to_cat['Column_Pole']['supercategory'] = 'civil-structure'
    dset.name_to_cat['Fence']['supercategory'] = 'civil-structure'
    dset.name_to_cat['Wall']['supercategory'] = 'civil-structure'
    dset.name_to_cat['Building']['supercategory'] = 'civil-structure'
    dset.name_to_cat['Archway']['supercategory'] = 'civil-structure'
    dset.name_to_cat['Bridge']['supercategory'] = 'civil-structure'
    dset.name_to_cat['Tunnel']['supercategory'] = 'civil-structure'

    dset.name_to_cat['TrafficCone']['supercategory'] = 'civil-notice'
    dset.name_to_cat['TrafficLight']['supercategory'] = 'civil-notice'
    dset.name_to_cat['LaneMkgsDriv']['supercategory'] = 'civil-notice'
    dset.name_to_cat['LaneMkgsNonDriv']['supercategory'] = 'civil-notice'
    dset.name_to_cat['SignSymbol']['supercategory'] = 'civil-notice'
    dset.name_to_cat['ParkingBlock']['supercategory'] = 'civil-notice'
    dset.name_to_cat['Misc_Text']['supercategory'] = 'civil-notice'

    dset.name_to_cat['Road']['supercategory'] = 'drive-way'
    dset.name_to_cat['RoadShoulder']['supercategory'] = 'drive-way'
    dset.name_to_cat['Sidewalk']['supercategory'] = 'walk-way'

    for cat in list(dset.cats.values()):
        parent = cat.get('supercategory', None)
        if parent is not None:
            if parent not in dset.name_to_cat:
                print('Missing parent = {!r}'.format(parent))
                dset.add_category(name=parent, supercategory=parent)

    if 0:
        graph = dset.category_graph()
        import graphid
        graphid.util.show_nx(graph)

    # Add in some hierarcy information
    if 0:
        for x in dset.name_to_cat:
            print("dset.name_to_cat[{!r}]['supercategory'] = 'object'".format(x))

    if 0:
        example_cat_aids = []
        for cat in dset.cats.values():
            cname = cat['name']
            aids = dset.index.cid_to_aids[dset.name_to_cat[cname]['id']]
            if len(aids):
                aid = ub.peek(aids)
                example_cat_aids.append(aid)
            else:
                print('No examples of cat = {!r}'.format(cat))

        import xdev
        import kwplot
        kwplot.autompl()
        for aid in xdev.InteractiveIter(example_cat_aids):
            print('aid = {!r}'.format(aid))
            ann = dset.anns[aid]
            cat = dset.cats[ann['category_id']]
            print('cat = {!r}'.format(cat))
            dset.show_image(aid=aid)
            xdev.InteractiveIter.draw()

        if 0:
            cname = 'CartLuggagePram'
            cname = 'ParkingBlock'
            cname = 'LaneMkgsDriv'
            aids = dset.index.cid_to_aids[dset.name_to_cat[cname]['id']]
            if len(aids):
                aid = ub.peek(aids)
                print('aid = {!r}'.format(aid))
                ann = dset.anns[aid]
                cat = dset.cats[ann['category_id']]
                print('cat = {!r}'.format(cat))
                dset.show_image(aid=aid)


def main():
    """
    Dump the paths to the coco file to stdout

    By default these will go to in the path:
        ~/.cache/netharn/camvid/camvid-master

    The four files will be:
        ~/.cache/netharn/camvid/camvid-master/camvid-full.mscoco.json
        ~/.cache/netharn/camvid/camvid-master/camvid-train.mscoco.json
        ~/.cache/netharn/camvid/camvid-master/camvid-vali.mscoco.json
        ~/.cache/netharn/camvid/camvid-master/camvid-test.mscoco.json
    """
    coco_dset = grab_coco_camvid()

    # Use the same train/test/vali splits used in segnet
    gid_subsets = grab_camvid_train_test_val_splits(coco_dset, mode='segnet')
    dpath = coco_dset.dataset['img_root']

    # Dump the full dataset
    fpath = join(dpath, 'camvid-full.mscoco.json')
    print(fpath)
    coco_dset.dump(open(fpath, 'w'))

    # Dump the train/vali/test splits
    for tag, gids in gid_subsets.items():
        subset = coco_dset.subset(gids)
        fpath = join(dpath, 'camvid-{}.mscoco.json'.format(tag))
        print(fpath)
        subset.dump(open(fpath, 'w'))

if __name__ == '__main__':
    """
    CommandLine:
        python -m netharn.data.grab_camvid
    """
    main()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
import torch.optim.lr_scheduler


class _LRScheduler(torch.optim.lr_scheduler._LRScheduler):
    """
    Fixes call to epoch 0 twice

    See:
        https://github.com/pytorch/pytorch/issues/8837
    """
    def __init__(self, optimizer, last_epoch=-1):
        if not isinstance(optimizer, torch.optim.lr_scheduler.Optimizer):
            raise TypeError('{} is not an Optimizer'.format(
                type(optimizer).__name__))
        self.optimizer = optimizer
        if last_epoch == -1:
            for group in optimizer.param_groups:
                group.setdefault('initial_lr', group['lr'])
        else:
            for i, group in enumerate(optimizer.param_groups):
                if 'initial_lr' not in group:
                    raise KeyError("param 'initial_lr' is not specified "
                                   "in param_groups[{}] when resuming an optimizer".format(i))
        self.base_lrs = list(map(lambda group: group['initial_lr'], optimizer.param_groups))
        self.last_epoch = last_epoch
        self.step(last_epoch)  # The major change is to remove the +1

    def get_lr(self):
        raise NotImplementedError

    def step(self, epoch=None):
        # epoch is really last epoch
        if epoch is None:
            epoch = self.last_epoch + 1
        self.last_epoch = epoch
        for param_group, lr in zip(self.optimizer.param_groups, self.get_lr()):
            param_group['lr'] = lr


class _LRScheduler2(_LRScheduler):
    """ Add a bit of extra functionality to the base torch LR scheduler """

    def current_lrs(self):
        lrs = [group['lr'] for group in self.optimizer.param_groups]
        return lrs


class ListedLR(_LRScheduler2):
    """
    Simple scheduler that simply sets the LR based on the epoch.

    Allows for hard-coded schedules for quick prototyping. Good for reproducing
    papers, but bad for experimentation.

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        points (dict): Mapping from epoch number to a learning rate.
            The epoch number indicates which epoch will be given a certain
            learning rate. Therefore if epoch 0 has a learning rate of 0.1 then
            that is what the LR will be set to BEFORE running epoch 0.
            Likewise if epoch 1 has a learning rate of 1.0, then AFTER epoch 0
            but BEFORE epoch 1, the learning rate will be set to 1.0.
        last_epoch (int): The index of last epoch. Default: -1.

    CommandLine:
        python ~/code/netharn/netharn/schedulers/listed.py ListedLR:1

    Example:
        >>> # Assuming optimizer has two groups.
        >>> import ubelt as ub
        >>> import netharn as nh
        >>> model = nh.models.ToyNet2d()
        >>> optimizer = torch.optim.SGD(model.parameters(), lr=0)
        >>> points = {0: .01, 2: .02, 3: .1, 6: .05, 9: .025}
        >>> self = ListedLR(optimizer, points)
        >>> lrs = [self._get_epoch_lr(epoch) for epoch in range(0, 11)]
        >>> print(list(ub.flatten(lrs)))
        [0.01, 0.01, 0.02, 0.1, 0.1, 0.1, 0.05, 0.05, 0.05, 0.025, 0.025]
        >>> assert self.current_lrs() == [0.01]
        >>> self = ListedLR(optimizer, points, interpolate=True)
        >>> lrs = [self._get_epoch_lr(epoch) for epoch in range(0, 11)]
        >>> print(ub.repr2(list(ub.flatten(lrs)), precision=3, nl=0))
        [0.010, 0.015, 0.020, 0.100, 0.083, 0.067, 0.050, 0.042, 0.033, 0.025, 0.025]

    Example:
        >>> # Assuming optimizer has two groups.
        >>> import ubelt as ub
        >>> import netharn as nh
        >>> model = nh.models.ToyNet2d()
        >>> optimizer = torch.optim.SGD(model.parameters(), lr=0)
        >>> points = {0: 1, 2: 2}
        >>> self = ListedLR(optimizer, points)
        >>> lrs = []
        >>> for i in range(5):
        >>>     lrs.append(self.get_lr())
        >>>     self.step()
        >>> print(list(ub.flatten(lrs)))
        [1, 1, 2, 2, 2]
    """

    def __init__(self, optimizer, points, interpolate=False,
                 last_epoch=-1):
        if not isinstance(points, dict):
            raise TypeError(points)
        self.interpolate = interpolate
        self.points = points

        # self.optimizer = optimizer
        # self.last_epoch = last_epoch

        # epochs where the lr changes
        self.key_epochs = sorted(self.points.keys())

        super(ListedLR, self).__init__(optimizer, last_epoch)

    def get_lr(self):
        lr = self._get_epoch_lr(self.last_epoch + 1)
        return lr

    def _get_epoch_lr(self, epoch):
        """ return lr based on the epoch """
        key_epochs = self.key_epochs
        points = self.points
        base_lrs = self.base_lrs

        assert epoch >= 0

        if epoch in key_epochs:
            prev_key_epoch = epoch
            next_key_epoch = epoch
        else:
            idx = np.searchsorted(key_epochs, epoch, 'left') - 1
            prev_key_epoch = key_epochs[idx]
            if idx < len(key_epochs) - 1:
                next_key_epoch = key_epochs[idx + 1]
            else:
                next_key_epoch = prev_key_epoch

        if self.interpolate:
            if next_key_epoch == prev_key_epoch:
                new_lr = points[next_key_epoch]
            else:
                prev_lr = points[next_key_epoch]
                next_lr = points[prev_key_epoch]

                alpha = (epoch - prev_key_epoch) / (next_key_epoch - prev_key_epoch)

                new_lr = alpha * prev_lr + (1 - alpha) * next_lr

            epoch_lrs = [new_lr for _ in base_lrs]
        else:
            if epoch < prev_key_epoch:
                epoch_lrs = base_lrs
            else:
                new_lr = points[prev_key_epoch]
                epoch_lrs = [new_lr for _ in base_lrs]
        return epoch_lrs


class Exponential(_LRScheduler2):
    """
    Decay learning rate by a factor of `gamma` every `stepsize` epochs.

    This class exists in torch, but lacks the stepsize parameter

    Example:
        >>> import ubelt as ub
        >>> import netharn as nh
        >>> model = nh.models.ToyNet2d()
        >>> optimizer = torch.optim.SGD(model.parameters(), lr=0.001)
        >>> self = Exponential(optimizer, gamma=0.01, stepsize=2)
        >>> rates = np.array([self._get_epoch_lr(i) for i in range(6)]).T[0]
        >>> target = np.array([1E-3, 1E-3, 1E-5, 1E-5, 1E-7, 1E-7])
        >>> assert np.allclose(target, rates)
    """
    def __init__(self, optimizer, gamma=0.1, stepsize=100):
        self.gamma = gamma
        self.stepsize = stepsize
        super(Exponential, self).__init__(optimizer)

    def get_lr(self):
        """
        If optimizer is specified, its learning rate is modified inplace.
        """
        new_lrs = self._get_epoch_lr(self.last_epoch + 1)
        return new_lrs

    def _get_epoch_lr(self, epoch):
        """ return lr based on the epoch """
        n_decays = epoch // self.stepsize
        new_lrs = [lr * (self.gamma ** n_decays) for lr in self.base_lrs]
        return new_lrs


class BatchLR(_LRScheduler2):
    __batchaware__ = True

    def __init__(self, optimizer, points, interpolate=False,
                 last_epoch=-1):
        if not isinstance(points, dict):
            raise TypeError(points)
        self.interpolate = interpolate
        self.points = points

        # self.optimizer = optimizer
        # self.last_epoch = last_epoch

        # epochs where the lr changes
        self.key_epochs = sorted(self.points.keys())

        super(BatchLR, self).__init__(optimizer, last_epoch)

        self.bx = 0

    def step_batch(self, bx=None):
        if bx is None:
            self.bx += 1
        else:
            self.bx = bx

    def step(self, *a, **kw):
        self.bx = 0
        return super(BatchLR, self).step(*a, **kw)

    step_epoch = step

    def get_lr(self):
        lr = self._get_epoch_lr(self.last_epoch + 1, self.bx)
        return lr

    def _get_epoch_lr(self, epoch, bx):
        """ return lr based on the epoch """
        key_epochs = self.key_epochs
        points = self.points
        base_lrs = self.base_lrs

        assert epoch >= 0

        if epoch in key_epochs:
            prev_key_epoch = epoch
            next_key_epoch = epoch
        else:
            idx = np.searchsorted(key_epochs, epoch, 'left') - 1
            prev_key_epoch = key_epochs[idx]
            if idx < len(key_epochs) - 1:
                next_key_epoch = key_epochs[idx + 1]
            else:
                next_key_epoch = prev_key_epoch

        if self.interpolate:
            if next_key_epoch == prev_key_epoch:
                new_lr = points[next_key_epoch]
            else:
                prev_lr = points[next_key_epoch]
                next_lr = points[prev_key_epoch]

                alpha = (epoch - prev_key_epoch) / (next_key_epoch - prev_key_epoch)

                new_lr = alpha * prev_lr + (1 - alpha) * next_lr

            epoch_lrs = [new_lr for _ in base_lrs]
        else:
            if epoch < prev_key_epoch:
                epoch_lrs = base_lrs
            else:
                new_lr = points[prev_key_epoch]
                epoch_lrs = [new_lr for _ in base_lrs]
        return epoch_lrs


if __name__ == '__main__':
    """
    CommandLine:
        python -m netharn.schedulers.listed all
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
Faster pure-python versions of sklearn functions that avoid expensive checks
and label rectifications. It is assumed that all labels are consecutive
non-negative integers.
"""
from scipy.sparse import coo_matrix
import numpy as np


def confusion_matrix(y_true, y_pred, n_labels=None, labels=None,
                     sample_weight=None):
    """
    faster version of sklearn confusion matrix that avoids the
    expensive checks and label rectification

    Runs in about 0.7ms

    Returns:
        ndarray: matrix where rows represent real and cols represent pred

    Example:
        >>> y_true = np.array([0, 0, 0, 0, 1, 1, 1, 0,  0, 1])
        >>> y_pred = np.array([0, 0, 0, 0, 0, 0, 0, 1,  1, 1])
        >>> confusion_matrix(y_true, y_pred, 2)
        array([[4, 2],
               [3, 1]])
        >>> confusion_matrix(y_true, y_pred, 2).ravel()
        array([4, 2, 3, 1])

    Benchmarks:
        import ubelt as ub
        y_true = np.random.randint(0, 2, 10000)
        y_pred = np.random.randint(0, 2, 10000)

        n = 1000
        for timer in ub.Timerit(n, bestof=10, label='py-time'):
            sample_weight = [1] * len(y_true)
            confusion_matrix(y_true, y_pred, 2, sample_weight=sample_weight)

        for timer in ub.Timerit(n, bestof=10, label='np-time'):
            sample_weight = np.ones(len(y_true), dtype=np.int)
            confusion_matrix(y_true, y_pred, 2, sample_weight=sample_weight)
    """
    if sample_weight is None:
        sample_weight = np.ones(len(y_true), dtype=np.int)
    if n_labels is None:
        n_labels = len(labels)
    CM = coo_matrix((sample_weight, (y_true, y_pred)),
                    shape=(n_labels, n_labels),
                    dtype=np.int64).toarray()
    return CM


def global_accuracy_from_confusion(cfsn):
    # real is rows, pred is columns
    n_ii = np.diag(cfsn)
    # sum over pred = columns = axis1
    t_i = cfsn.sum(axis=1)
    global_acc = n_ii.sum() / t_i.sum()
    return global_acc


def class_accuracy_from_confusion(cfsn):
    # real is rows, pred is columns
    n_ii = np.diag(cfsn)
    # sum over pred = columns = axis1
    t_i = cfsn.sum(axis=1)
    per_class_acc = (n_ii / t_i).mean()
    class_acc = np.nan_to_num(per_class_acc).mean()
    return class_acc
